"""Interface adapters to scikit-survival Cox PH model."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

from skpro.survival.adapters.sksurv import _SksurvAdapter
from skpro.survival.base import BaseSurvReg


class CoxPHSkSurv(_SksurvAdapter, BaseSurvReg):
    """Cox proportional hazards, from scikit-survival.

    Direct interface to ``sksurv.linear_model.CoxPHSurvivalAnalysis``, by ``sebp``.

    Parameters
    ----------
    alpha : float, or ndarray of shape (n_features,), optional, default: 0
        Regularization parameter for ridge regression penalty.
        If a single float, the same penalty is used for all features.
        If an array, there must be one penalty for each feature.
        If you want to include a subset of features without penalization,
        set the corresponding entries to 0.

    ties : {'breslow', 'efron'}, optional, default: 'breslow'
        The method to handle tied event times. If there are
        no tied event times all the methods are equivalent.

    n_iter : int, optional, default: 100
        Maximum number of iterations.

    tol : float, optional, default: 1e-9
        Convergence criteria. Convergence is based on the negative log-likelihood:

        |1 - (new neg. log-likelihood / old neg. log-likelihood) | < tol

    verbose : int, optional, default: 0
        Specifies the amount of additional debug information
        during optimization.

    Attributes
    ----------
    coef_ : ndarray, shape = (n_features,)
        Coefficients of the model

    cum_baseline_hazard_ : ``sksurv.functions.StepFunction``
        Estimated baseline cumulative hazard function.

    baseline_survival_ : ``sksurv.functions.StepFunction``
        Estimated baseline survival function.

    unique_times_ : array of shape = (n_unique_times,)
        Unique time points.
    """

    _tags = {"authors": ["sebp", "fkiraly"]}  # sebp credit for interfaced estimator

    def __init__(self, alpha=0, ties="breslow", n_iter=100, tol=1e-9, verbose=0):
        self.alpha = alpha
        self.ties = ties
        self.n_iter = n_iter
        self.tol = tol
        self.verbose = verbose

        super().__init__()

    def _get_sksurv_class(self):
        """Getter of the sksurv class to be used for the adapter."""
        from sksurv.linear_model import CoxPHSurvivalAnalysis as _CoxPH

        return _CoxPH

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}

        params2 = {
            "alpha": 0.1,
            "ties": "efron",
            "n_iter": 99,
            "tol": 1e-7,
            "verbose": 1,
        }

        return [params1, params2]



"""Probabilistic regression by bootstrap."""

__author__ = ["fkiraly"]
__all__ = ["EnbpiRegressor"]

import numpy as np
import pandas as pd
from sklearn import clone
from sklearn.utils import check_random_state

from skpro.distributions.empirical import Empirical
from skpro.regression.base import BaseProbaRegressor
from skpro.utils.numpy import flatten_to_1D_if_colvector
from skpro.utils.sklearn import prep_skl_df


class EnbpiRegressor(BaseProbaRegressor):
    r"""EnbPI regressor, aka Jackknife+-after-bootstrap, conformal intervals.

    Wraps an ``sklearn`` regressor and turns it into an ``skpro`` regressor
    with access to all probabilistic prediction methods.

    Follows the original algorithms in [1]_ and [2]_.
    In ``predict_proba``, a distribution
    implied by the quantile predictions is returned, an ``Empirical`` distribution.

    Fits ``n_estimators`` clones of a tabular ``sklearn`` regressor on
    datasets which are bootstrap sub-samples, i.e.,
    independent row samples with replacement.

    The clones are aggregated to predict quantiles of the target distribution,
    following the original algorithms in [1]_ and [2]_.

    The parameters in the reference [2]_ are mapped to the parameters of the
    estimator as follows: :math:`\mathcal{A}` is ``estimator``, :math:`B` is
    ``n_bootstrap_samples``, :math:`x_i, i = 1, \dots, T` are the rows of
    ``X`` in ``fit``, :math:`y_i, i = 1, \dots, T` are the rows of ``y`` in ``fit``,
    :math:`x_i, i = T+1, \dots, T + T_1` are the rows of ``X`` in ``predict_interval``,
    :math:`\phi` is ``agg_fun``.

    The :math:`C_{T, t}^{\phi, \alpha}(x_t), t = T+1, \dots, T + T_1` are encoded
    in the prediction object returned by ``predict_interval`` or ``predict_proba``.

    Parameters
    ----------
    estimator : sklearn regressor
        regressor to use in the bootstrap
    n_bootstrap_samples : int, default=100
        The number of bootstrap samples drawn
        If int, then indicates number of instances precisely
        Note: this is not the same as the size of each bootstrap sample.
        The size of the bootstrap sample is always equal to X.
    agg_fun : str or callable, default="mean"
        function to aggregate the predictions of the bootstrap samples
        If str, must be one of "mean" or "median", meaning np.mean or np.median
        If callable, must be a function of signature (n, m) -> m, or (n) -> 1
    symmetrize : bool, default=True
        whether to symmetrize the prediction intervals and predictive quantiles
        Default = True leads to the original algorithm in [2]_.
        If False, the conformalized sample and the prediction intervals are not
        symmetrized.
    replace : bool, default=True
        whether to sample the bootstrap sample with replacement or without replacement.
        Default = True leads to the algorithm in [2]_.
    random_state : int, RandomState instance or None, optional (default=None)
        If int, ``random_state`` is the seed used by the random number generator;
        If ``RandomState`` instance, ``random_state`` is the random number generator;
        If None, the random number generator is the ``RandomState`` instance used
        by ``np.random``.

    Attributes
    ----------
    estimators_ : list of of skpro regressors
        clones of regressor in `estimator` fitted in the ensemble

    References
    ----------
    .. [1] Byol Kim, Chen Xu, and Rina Foygel Barber (2020).
      Predictive Inference Is Free with the Jackknife+-after-Bootstrap.
      Advances in Neural Information Processing Systems 33, 2020.
    .. [2] Xu, Chen and Yao Xie (2021).
      Conformal prediction interval for dynamic time-series.
      The Proceedings of the 38th International Conference on Machine Learning,
      PMLR 139, 2021.

    Examples
    --------
    >>> from skpro.regression.enbpi import EnbpiRegressor
    >>> from sklearn.linear_model import LinearRegression
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.model_selection import train_test_split
    >>>
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)
    >>>
    >>> reg_tabular = LinearRegression()
    >>>
    >>> reg_proba = EnbpiRegressor(reg_tabular)
    >>> reg_proba.fit(X_train, y_train)
    EnbpiRegressor(...)
    >>> y_pred = reg_proba.predict_proba(X_test)
    """

    _tags = {
        "authors": ["fkiraly", "hamrel-cxu"],
        "capability:missing": True,
    }

    def __init__(
        self,
        estimator,
        n_bootstrap_samples=100,
        agg_fun="mean",
        symmetrize=True,
        replace=True,
        random_state=None,
    ):
        self.estimator = estimator
        self.n_bootstrap_samples = n_bootstrap_samples
        self.agg_fun = agg_fun
        self.symmetrize = symmetrize
        self.replace = replace
        self.random_state = random_state
        self._random_state = check_random_state(random_state)

        super().__init__()

        # todo: find the equivalent tag in sklearn for missing data handling
        # tags_to_clone = ["capability:missing"]
        # self.clone_tags(estimator, tags_to_clone)

        aggfun_dict = {
            "mean": np.nanmean,
            "median": np.nanmedian,
        }
        if agg_fun is None:
            self._agg_fun = np.nanmean
        elif agg_fun in aggfun_dict:
            self._agg_fun = aggfun_dict[agg_fun]
        else:
            self._agg_fun = agg_fun

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        estimator = self.estimator
        n_bootstrap_samples = self.n_bootstrap_samples
        np.random.seed(self.random_state)

        inst_ix = X.index
        n = len(inst_ix)

        self.estimators_ = []
        self._cols = y.columns

        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)

        y_pred_bs = np.ones((n_bootstrap_samples,) + y.shape) * np.nan
        bs_vs_ix = np.zeros((n, n_bootstrap_samples))

        for i in range(n_bootstrap_samples):
            esti = clone(estimator)
            row_iloc = pd.RangeIndex(n)
            row_ss = _random_ss_ix(
                row_iloc, size=n, replace=self.replace, random_state=self._random_state
            )
            inst_ix_i = inst_ix[row_ss]
            bs_vs_ix[np.unique(row_ss), i] = 1

            Xi = X.loc[inst_ix_i]
            Xi = Xi.reset_index(drop=True)

            yi = y.loc[inst_ix_i].values
            yi = flatten_to_1D_if_colvector(yi)

            self.estimators_ += [esti.fit(Xi, yi)]

            y_pred = esti.predict(X)
            y_pred_bs[[i]] = _coerce_numpy2d(y_pred)

        y_pred_insample = np.ones(y.shape) * np.nan

        for i in range(n):
            y_pred_insample[[i], :] = self._pred_phi_sans_i(
                y_pred_bs[:, [i], :], i, bs_vs_ix
            )

        errs = y.values - y_pred_insample
        if self.symmetrize:
            errs = np.abs(errs)

        self._errs = errs
        self._bs_vs_ix = bs_vs_ix
        self._n_train = n

        return self

    def _pred_phi_sans_i(self, y_preds, i, bs_vs_ix):
        # y_preds - (n_bootstrap_samples, n_samples, n_vars)
        # bs_vs_ix - (n_train_samples, n_bootstrap_samples)
        # output: (n_samples, n_vars)
        bs_wo_i = (bs_vs_ix[i] == 0).flatten()
        y_preds_wo_i = y_preds[bs_wo_i]
        return self._agg_preds(y_preds_wo_i)

    def _agg_preds(self, preds):
        """Aggregate predictions of bootstrap samples.

        Handles NaNs in the aggregated predictions
        by replacing them with the aggregated value from the same column
        (index of target variable), after the aggregation.

        Parameters
        ----------
        preds : np.ndarray, shape (n_bootstrap_samples, n_samples, n_targets)
            predictions of bootstrap samples

        Returns
        -------
        agg_preds : np.ndarray, shape (n_samples, n_targets)
            aggregated predictions
        """
        spl = self._agg_fun(preds, axis=0)
        nans = np.isnan(spl)
        if nans.any():
            for j in range(spl.shape[1]):
                colj = spl[:, j]
                colj_agg = self._agg_fun(colj)
                colj[nans[:, j]] = colj_agg
                spl[:, j] = colj
        return spl

    def _predict_proba(self, X) -> np.ndarray:
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        cols = self._cols
        n_cols = len(cols)  # number of targets
        n_est = self.n_bootstrap_samples
        n_train = self._n_train
        estimators_ = self.estimators_
        bs_vs_ix = self._bs_vs_ix
        _errs = self._errs

        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)
        n = len(X)

        y_preds = np.zeros((n_est, n, n_cols))

        for i, est in enumerate(estimators_):
            y_preds[i] = _coerce_numpy2d(est.predict(X))

        # y_preds_agg - (n_train, n, len(cols))
        y_preds_agg = np.zeros((n_train, n, n_cols))
        for i in range(n_train):
            y_preds_agg[[i], :, :] = self._pred_phi_sans_i(y_preds, i, bs_vs_ix)

        if self.symmetrize:
            errs = np.concatenate([_errs, -_errs], axis=0)
            y_preds_rep = np.tile(y_preds_agg, (2, 1, 1))
        else:
            errs = _errs
            y_preds_rep = y_preds_agg

        errs = np.expand_dims(errs, axis=1)

        n_emp_spl = len(errs)
        emp_df_vals = y_preds_rep + errs
        emp_df_vals = np.reshape(emp_df_vals, (n_emp_spl * n, n_cols))

        emp_ix = pd.MultiIndex.from_product([range(n_emp_spl), X.index])

        spl_df = pd.DataFrame(emp_df_vals, index=emp_ix, columns=cols)
        y_proba = Empirical(spl_df)
        return y_proba

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.linear_model import LinearRegression

        params1 = {"estimator": LinearRegression()}
        params2 = {
            "estimator": LinearRegression(),
            "n_bootstrap_samples": 10,
        }
        params3 = {
            "estimator": LinearRegression(),
            "n_bootstrap_samples": 12,
            "agg_fun": "median",
            "symmetrize": False,
        }

        return [params1, params2, params3]


def _random_ss_ix(ix, size, replace=True, random_state=None):
    """Randomly uniformly sample indices from a list of indices."""
    if random_state is None:
        random_state = np.random.RandomState()

    a = range(len(ix))
    ixs = ix[random_state.choice(a, size=size, replace=replace)]
    return ixs


def _coerce_numpy2d(x):
    if x.ndim == 1:
        x = x.reshape(-1, 1)
    return x



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Common abstraction utilities for parallelization backends.

New parallelization or iteration backends can be added easily as follows:

* Add a new backend name to ``backend_dict``, syntax is
  backend_name: backend_type, where backend_type collects backend options,
  e.g., multiple options for a single parallelization backend.
* Add a new function to ``para_dict``, should have name
  ``_parallelize_<backend_name>`` and take the same arguments as
  ``_parallelize_none``. Ensure that ``backend`` and ``backend_params`` are arguments,
  even if there is only one backend option, or no additional parameters.
* add the backend string in the docstring of parallelize, and any downstream
  functions that use ``parallelize`` and expose the backend parameter an argument
"""

__author__ = ["fkiraly"]


def parallelize(fun, iter, meta=None, backend=None, backend_params=None):
    """Parallelize loop over iter via backend.

    Executes ``fun(x, meta)`` in parallel for ``x`` in ``iter``,
    and returns the results as a list in the same order as ``iter``.

    Uses the iteration or parallelization backend specified by ``backend``.

    Parameters
    ----------
    fun : callable
        function to be executed in parallel
    iter : iterable
        iterable over which to parallelize
    meta : dict, optional
        variables to be passed to fun
    backend : str, optional
        backend to use for parallelization, one of

        - "None": executes loop sequentally, simple list comprehension
        - "loky", "multiprocessing" and "threading": uses ``joblib`` ``Parallel`` loops
        - "joblib": custom and 3rd party ``joblib`` backends, e.g., ``spark``
        - "dask": uses ``dask``, requires ``dask`` package in environment
        - "dask_lazy": same as ``"dask"``, but returns delayed object instead of list

    backend_params : dict, optional
        additional parameters passed to the backend as config.
        Valid keys depend on the value of ``backend``:

        - "None": no additional parameters, ``backend_params`` is ignored
        - "loky", "multiprocessing" and "threading": default ``joblib`` backends
          any valid keys for ``joblib.Parallel`` can be passed here, e.g., ``n_jobs``,
          with the exception of ``backend`` which is directly controlled by ``backend``.
          If ``n_jobs`` is not passed, it will default to ``-1``, other parameters
          will default to ``joblib`` defaults.
        - "joblib": custom and 3rd party ``joblib`` backends, e.g., ``spark``.
          any valid keys for ``joblib.Parallel`` can be passed here, e.g., ``n_jobs``,
          ``backend`` must be passed as a key of ``backend_params`` in this case.
          If ``n_jobs`` is not passed, it will default to ``-1``, other parameters
          will default to ``joblib`` defaults.
        - "dask": any valid keys for ``dask.compute`` can be passed, e.g., ``scheduler``
    """
    if meta is None:
        meta = {}
    if backend is None:
        backend = "None"
    if backend_params is None:
        backend_params = {}

    backend_name = backend_dict[backend]
    para_fun = para_dict[backend_name]

    ret = para_fun(
        fun=fun, iter=iter, meta=meta, backend=backend, backend_params=backend_params
    )
    return ret


backend_dict = {
    "None": "none",
    "loky": "joblib",
    "multiprocessing": "joblib",
    "threading": "joblib",
    "joblib": "joblib",
    "dask": "dask",
    "dask_lazy": "dask",
}
para_dict = {}


def _parallelize_none(fun, iter, meta, backend, backend_params):
    """Execute loop via simple sequential list comprehension."""
    ret = [fun(x, meta=meta) for x in iter]
    return ret


para_dict["none"] = _parallelize_none


def _parallelize_joblib(fun, iter, meta, backend, backend_params):
    """Parallelize loop via joblib Parallel."""
    from joblib import Parallel, delayed

    par_params = backend_params.copy()
    if "backend" not in par_params:
        # if user selects custom joblib backend but does not specify backend explicitly,
        # raise a ValueError
        if backend == "joblib":
            raise ValueError(
                '"joblib" was selected as first layer parallelization backend, '
                "but no backend string was "
                '"passed in the backend parameters dict, e.g., "spark". '
                "Please specify a backend to joblib as a key-value pair "
                "in the backend_params arg or the backend:parallel:params config "
                'when using "joblib". '
                'For clarity, "joblib" should only be used for two-layer '
                "backend dispatch, where the first layer is joblib, "
                "and the second layer is a custom backend of joblib, e.g., spark."
                "For first-party joblib backends, please use the backend string "
                'of sktime directly, e.g., by specifying "multiprocessing" or "loky".'
            )
        # in all other cases, we ensure the backend parameter is one of
        # "loky", "multiprocessing" or "threading", as passed via backend
        else:
            par_params["backend"] = backend
    elif backend != "joblib":
        par_params["backend"] = backend

    if "n_jobs" not in par_params:
        par_params["n_jobs"] = -1

    ret = Parallel(**par_params)(delayed(fun)(x, meta=meta) for x in iter)
    return ret


para_dict["joblib"] = _parallelize_joblib


def _parallelize_dask(fun, iter, meta, backend, backend_params):
    """Parallelize loop via dask."""
    from dask import compute, delayed

    lazy = [delayed(fun)(x, meta=meta) for x in iter]
    if backend == "dask":
        return compute(*lazy, **backend_params)
    else:
        return lazy


para_dict["dask"] = _parallelize_dask


def _get_parallel_test_fixtures():
    """Return fixtures for parallelization tests.

    Returns a list of parameter fixtures, where each fixture
    is a dict with keys "backend" and "backend_params".
    """
    from skbase.utils.dependencies import _check_soft_dependencies

    fixtures = []

    # test no parallelization
    fixtures.append({"backend": "None", "backend_params": {}})

    # test joblib backends
    for backend in ["loky", "multiprocessing", "threading"]:
        fixtures.append({"backend": backend, "backend_params": {}})
        fixtures.append({"backend": backend, "backend_params": {"n_jobs": 2}})
        fixtures.append({"backend": backend, "backend_params": {"n_jobs": -1}})

    # test dask backends
    if _check_soft_dependencies("dask", severity="none"):
        fixtures.append({"backend": "dask", "backend_params": {}})
        fixtures.append({"backend": "dask", "backend_params": {"scheduler": "sync"}})

    return fixtures



"""Legacy module: test utils."""
# LEGACY MODULE - TODO: remove or refactor

import numpy as np


def assert_close_prediction(y_hat, y_true, fraction=0.75, within=0.25):
    """Check that defined fraction of predictions lies in a certain tolerance.

    Parameters
    ----------
    y_hat   Predictions
    y_true  True values
    fraction Fraction of close values
    within  Relative tolerance to assume when comparing the values

    Raises
    ------
    AssertionError
    """
    predictions_within_tolerance = np.count_nonzero(
        np.isclose(y_hat, y_true, rtol=within)
    )
    target = len(y_true) * fraction

    assert predictions_within_tolerance > target



"""Adapters for probabilistic regressors, towards sklearn."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

from skpro.regression.adapters.ngboost._ngboost_proba import NGBoostAdapter

__all__ = ["NGBoostAdapter"]



"""Tests Generalized Linear Model regressor."""

import pandas as pd
import pytest

from skpro.regression.linear import GLMRegressor
from skpro.tests.test_switch import run_test_for_class


@pytest.mark.skipif(
    not run_test_for_class(GLMRegressor),
    reason="run test only if softdeps are present and incrementally (if requested)",
)
def test_glm_simple_use():
    """Test simple use of GLM regressor."""
    from sklearn.datasets import load_diabetes
    from sklearn.model_selection import train_test_split

    X, y = load_diabetes(return_X_y=True, as_frame=True)
    y = pd.DataFrame(y)
    X = X.iloc[:200]
    y = y.iloc[:200]
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

    glm_reg = GLMRegressor()
    glm_reg.fit(X_train, y_train)
    y_pred = glm_reg.predict(X_test)
    y_pred_proba = glm_reg.predict_proba(X_test)

    assert y_pred.shape == y_test.shape
    assert y_pred_proba.shape == y_test.shape


@pytest.mark.skipif(
    not run_test_for_class(GLMRegressor),
    reason="run test only if softdeps are present and incrementally (if requested)",
)
def test_glm_with_offset_exposure():
    """Test GLM with offset_var and exposure_var parameters."""
    import numpy as np
    from sklearn.datasets import load_diabetes
    from sklearn.model_selection import train_test_split

    X, y = load_diabetes(return_X_y=True, as_frame=True)
    y = pd.DataFrame(y)
    X = X.iloc[:200]
    y = y.iloc[:200]
    X["off"] = np.ones(X.shape[0]) * 2.1
    X["exp"] = np.arange(1, X.shape[0] + 1)
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

    glm_reg = GLMRegressor(
        family="Normal", link="Log", offset_var="off", exposure_var=-1
    )
    glm_reg.fit(X_train, y_train)
    y_pred = glm_reg.predict(X_test)
    y_pred_proba = glm_reg.predict_proba(X_test)

    assert y_pred.shape == y_test.shape
    assert y_pred_proba.shape == y_test.shape



"""Tests for Harell's C-index."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

import pandas as pd
import pytest


@pytest.mark.parametrize("concordant", [True, False])
@pytest.mark.parametrize("pass_c", ["True", "False", "None"])
@pytest.mark.parametrize("normalization", ["overall", "index"])
def test_charrell_logic(concordant, pass_c, normalization):
    """Test the logic of the Harrell's C-index metric.

    Parameters
    ----------
    concordant : bool, optional, default=True
        If True, the test examples are fully concordant.
        If False, the test examples are fully discordant.
    pass_c : bool, optional, default=True
        If True, the ``C_true`` argument is passed to the metric, with censoring data.
        If None, the ``C_true`` argument is passed to the metric, with value None.
        If False, the ``C_true`` argument is not passed to the metric.
    normalization : str, optional, default="overall"
        The normalization method for the metric.
    """
    from skpro.distributions import Normal
    from skpro.metrics.survival._c_harrell import ConcordanceHarrell

    # examples are constructed to be fully concordant or discordant,
    # depending on the value of `concordant`
    y_true = pd.DataFrame({"a": [1, 2, 3, 4], "b": [5, 4, 3, 2]})
    c_true = pd.DataFrame({"a": [1, 0, 1, 0], "b": [0, 1, 0, 1]})
    y_pred_mean = pd.DataFrame({"a": [2, 3, 4, 5], "b": [6, 5, 4, 3]})

    if not concordant:
        y_pred_mean = -y_pred_mean
    y_pred = Normal(y_pred_mean, sigma=1, columns=pd.Index(["a", "b"]))

    # evaluate the metric
    metric = ConcordanceHarrell(normalization=normalization, tie_score=int(concordant))
    metric_args = {"y_true": y_true, "y_pred": y_pred}
    if pass_c == "True":
        metric_args["C_true"] = c_true
    elif pass_c == "None":
        metric_args["C_true"] = c_true

    res = metric(**metric_args)
    res_by_index = metric.evaluate_by_index(**metric_args)

    assert res_by_index.shape == y_true.shape

    # test assumptions
    # if concordant, the result should be 1
    # if discordant, the result should be 0
    assert res == concordant

    if normalization == "index":
        assert (res_by_index == concordant).all().all()



"""Interface adapters to scikit-survival gradient boosting models."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

from skpro.survival.adapters.sksurv import _SksurvAdapter
from skpro.survival.base import BaseSurvReg


class SurvGradBoostSkSurv(_SksurvAdapter, BaseSurvReg):
    """Gradient-boosted survival trees with proportional hazards loss, from sksurv.

    Direct interface to ``sksurv.ensemble.boosting.GradientBoostingSurvivalAnalysis``.

    In each stage, a regression tree is fit on the negative gradient
    of the loss function.

    For more details on gradient boosting see [1]_ and [2]_. If `loss='coxph'`,
    the partial likelihood of the proportional hazards model is optimized as
    described in [3]_. If `loss='ipcwls'`, the accelerated failure time model with
    inverse-probability of censoring weighted least squares error is optimized as
    described in [4]_. When using a non-zero `dropout_rate`, regularization is
    applied during training following [5]_.

    Parameters
    ----------
    loss : {'coxph', 'squared', 'ipcwls'}, optional, default: 'coxph'
        loss function to be optimized. 'coxph' refers to partial likelihood loss
        of Cox's proportional hazards model. The loss 'squared' minimizes a
        squared regression loss that ignores predictions beyond the time of censoring,
        and 'ipcwls' refers to inverse-probability of censoring weighted least squares
        error.

    learning_rate : float, optional, default: 0.1
        learning rate shrinks the contribution of each tree by ``learning_rate``.
        There is a trade-off between ``learning_rate`` and ``n_estimators``.
        Values must be in the range ``[0.0, inf)``.

    n_estimators : int, default: 100
        The number of regression trees to create. Gradient boosting
        is fairly robust to over-fitting so a large number usually
        results in better performance.
        Values must be in the range ``[1, inf)``.

    subsample : float, optional, default: 1.0
        The fraction of samples to be used for fitting the individual base
        learners. If smaller than 1.0 this results in Stochastic Gradient
        Boosting. ``subsample`` interacts with the parameter `n_estimators`.
        Choosing ``subsample < 1.0`` leads to a reduction of variance
        and an increase in bias.
        Values must be in the range ``(0.0, 1.0]``.

    criterion : string, optional, "squared_error" or "friedman_mse" (default)
        The function to measure the quality of a split. Supported criteria
        are "friedman_mse" for the mean squared error with improvement
        score by Friedman, "squared_error" for mean squared error.
        The default value of "friedman_mse" is
        generally the best as it can provide a better approximation in
        some cases.

    min_samples_split : int or float, optional, default: 2
        The minimum number of samples required to split an internal node:

        - If int, values must be in the range ``[2, inf)``.
        - If float, values must be in the range ``(0.0, 1.0]`` and ``min_samples_split``
          will be ``ceil(min_samples_split * n_samples)``.

    min_samples_leaf : int or float, default: 1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, values must be in the range ``[1, inf)``.
        - If float, values must be in the range ``(0.0, 1.0)`` and ``min_samples_leaf``
          will be ``ceil(min_samples_leaf * n_samples)``.

    min_weight_fraction_leaf : float, optional, default: 0.
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when `sample_weight` is not provided.
        Values must be in the range ``[0.0, 0.5]``.

    max_depth : int or None, optional, default: 3
        Maximum depth of the individual regression estimators. The maximum
        depth limits the number of nodes in the tree. Tune this parameter
        for best performance; the best value depends on the interaction
        of the input variables. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        `min_samples_split` samples.
        If int, values must be in the range ``[1, inf)``.

    min_impurity_decrease : float, optional, default: 0.
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

    max_features : int, float, string or None, optional, default: None
        The number of features to consider when looking for the best split:

        - If int, values must be in the range ``[1, inf)``.
        - If float, values must be in the range ``(0.0, 1.0]`` and the features
          considered at each split will
          be ``max(1, int(max_features * n_features_in_))``.
        - If 'sqrt', then ``max_features=sqrt(n_features)``.
        - If 'log2', then ``max_features=log2(n_features)``.
        - If None, then ``max_features=n_features``.

        Choosing ``max_features < n_features`` leads to a reduction of variance
        and an increase in bias.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_leaf_nodes : int or None, optional, default: None
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        Values must be in the range `[2, inf)`.
        If ``None``, then unlimited number of leaf nodes.

    validation_fraction : float, default: 0.1
        The proportion of training data to set aside as validation set for
        early stopping. Values must be in the range ``(0.0, 1.0)``.
        Only used if ``n_iter_no_change`` is set to an integer.

    n_iter_no_change : int, default: None
        ``n_iter_no_change`` is used to decide if early stopping will be used
        to terminate training when validation score is not improving. By
        default it is set to None to disable early stopping. If set to a
        number, it will set aside ``validation_fraction`` size of the training
        data as validation and terminate training when validation score is not
        improving in all of the previous ``n_iter_no_change`` numbers of
        iterations. The split is stratified.
        Values must be in the range ``[1, inf)``.

    tol : float, default: 1e-4
        Tolerance for the early stopping. When the loss is not improving
        by at least tol for ``n_iter_no_change`` iterations (if set to a
        number), the training stops.
        Values must be in the range ``[0.0, inf)``.

    dropout_rate : float, optional, default: 0.0
        If larger than zero, the residuals at each iteration are only computed
        from a random subset of base learners. The value corresponds to the
        percentage of base learners that are dropped. In each iteration,
        at least one base learner is dropped. This is an alternative regularization
        to shrinkage, i.e., setting ``learning_rate < 1.0``.
        Values must be in the range ``[0.0, 1.0)``.

    ccp_alpha : non-negative float, optional, default: 0.0.
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed.
        Values must be in the range ``[0.0, inf)``.

    verbose : int, default: 0
        Enable verbose output. If 1 then it prints progress and performance
        once in a while (the more trees the lower the frequency). If greater
        than 1 then it prints progress and performance for every tree.
        Values must be in the range ``[0, inf)``.

    random_state : int seed, RandomState instance, or None, default: None
        Controls the random seed given to each Tree estimator at each
        boosting iteration.
        In addition, it controls the random permutation of the features at
        each split.
        It also controls the random splitting of the training data to obtain a
        validation set if ``n_iter_no_change`` is not None.
        Pass an int for reproducible output across multiple function calls.

    Attributes
    ----------
    n_estimators_ : int
        The number of estimators as selected by early stopping (if
        ``n_iter_no_change`` is specified). Otherwise it is set to
        ``n_estimators``.

    feature_importances_ : ndarray, shape = (n_features,)
        The feature importances (the higher, the more important the feature).

    estimators_ : ndarray of DecisionTreeRegressor, shape = (n_estimators, 1)
        The collection of fitted sub-estimators.

    train_score_ : ndarray, shape = (n_estimators,)
        The i-th score ``train_score_[i]`` is the deviance (= loss) of the
        model at iteration ``i`` on the in-bag sample.
        If ``subsample == 1`` this is the deviance on the training data.

    oob_improvement_ : ndarray, shape = (n_estimators,)
        The improvement in loss (= deviance) on the out-of-bag samples
        relative to the previous iteration.
        ``oob_improvement_[0]`` is the improvement in
        loss of the first stage over the ``init`` estimator.

    unique_times_ : array of shape = (n_unique_times,)
        Unique time points.

    References
    ----------
    .. [1] J. H. Friedman, "Greedy function approximation: A gradient boosting machine,"
           The Annals of Statistics, 29(5), 1189–1232, 2001.
    .. [2] J. H. Friedman, "Stochastic gradient boosting,"
           Computational Statistics & Data Analysis, 38(4), 367–378, 2002.
    .. [3] G. Ridgeway, "The state of boosting,"
           Computing Science and Statistics, 172–181, 1999.
    .. [4] Hothorn, T., Bühlmann, P., Dudoit, S., Molinaro, A., van der Laan, M. J.,
           "Survival ensembles", Biostatistics, 7(3), 355-73, 2006.
    .. [5] K. V. Rashmi and R. Gilad-Bachrach,
           "DART: Dropouts meet multiple additive regression trees,"
           in 18th International Conference on Artificial Intelligence and Statistics,
           2015, 489–497.
    """

    _tags = {
        "authors": ["sebp", "fkiraly"],  # sebp credit for interfaced estimator=
    }

    def __init__(
        self,
        loss="coxph",
        learning_rate=0.1,
        n_estimators=100,
        subsample=1.0,
        criterion="friedman_mse",
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_depth=3,
        min_impurity_decrease=0.0,
        max_features=None,
        max_leaf_nodes=None,
        validation_fraction=0.1,
        n_iter_no_change=None,
        tol=1e-4,
        dropout_rate=0.0,
        ccp_alpha=0.0,
        verbose=0,
        random_state=None,
    ):
        self.loss = loss
        self.learning_rate = learning_rate
        self.n_estimators = n_estimators
        self.subsample = subsample
        self.criterion = criterion
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_depth = max_depth
        self.min_impurity_decrease = min_impurity_decrease
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.validation_fraction = validation_fraction
        self.n_iter_no_change = n_iter_no_change
        self.tol = tol
        self.dropout_rate = dropout_rate
        self.ccp_alpha = ccp_alpha
        self.verbose = verbose
        self.random_state = random_state

        super().__init__()

    def _get_sksurv_class(self):
        """Getter of the sksurv class to be used for the adapter."""
        from sksurv.ensemble.boosting import GradientBoostingSurvivalAnalysis

        return GradientBoostingSurvivalAnalysis

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}

        params2 = {
            "learning_rate": 0.2,
            "n_estimators": 99,
            "subsample": 0.9,
            "criterion": "squared_error",
            "min_samples_split": 3,
            "min_samples_leaf": 2,
            "max_depth": 4,
            "max_features": "sqrt",
            "tol": 1e-3,
            "dropout_rate": 0.01,
        }

        return [params1, params2]


class SurvGradBoostCompSkSurv(_SksurvAdapter, BaseSurvReg):
    r"""Survival Gradient boosting component-wise least squares, from sksurv.

    Direct interface to ``ComponentwiseGradientBoostingSurvivalAnalysis``
    from ``sksurv.ensemble.boosting``.

    Parameters
    ----------
    loss : {'coxph', 'squared', 'ipcwls'}, optional, default: 'coxph'
        loss function to be optimized. 'coxph' refers to partial likelihood loss
        of Cox's proportional hazards model. The loss 'squared' minimizes a
        squared regression loss that ignores predictions beyond the time of censoring,
        and 'ipcwls' refers to inverse-probability of censoring weighted least squares
        error.

    learning_rate : float, optional, default: 0.1
        learning rate shrinks the contribution of each base learner by `learning_rate`.
        There is a trade-off between `learning_rate` and `n_estimators`.
        Values must be in the range ``[0.0, inf)``.

    n_estimators : int, default: 100
        The number of boosting stages to perform. Gradient boosting
        is fairly robust to over-fitting so a large number usually
        results in better performance.
        Values must be in the range ``[1, inf)``.

    subsample : float, optional, default: 1.0
        The fraction of samples to be used for fitting the individual base
        learners. If smaller than 1.0 this results in Stochastic Gradient
        Boosting. ``subsample`` interacts with the parameter ``n_estimators``.
        Choosing ``subsample < 1.0`` leads to a reduction of variance
        and an increase in bias.
        Values must be in the range ``(0.0, 1.0]``.

    dropout_rate : float, optional, default: 0.0
        If larger than zero, the residuals at each iteration are only computed
        from a random subset of base learners. The value corresponds to the
        percentage of base learners that are dropped. In each iteration,
        at least one base learner is dropped. This is an alternative regularization
        to shrinkage, i.e., setting ``learning_rate < 1.0``.
        Values must be in the range ``[0.0, 1.0)``.

    random_state : int seed, RandomState instance, or None, default: None
        The seed of the pseudo random number generator to use when
        shuffling the data.

    verbose : int, default: 0
        Enable verbose output. If 1 then it prints progress and performance
        once in a while.
        Values must be in the range ``[0, inf)``.

    Attributes
    ----------
    coef_ : array, shape = (n_features + 1,)
        The aggregated coefficients. The first element `coef\_[0]` corresponds
        to the intercept. If loss is `coxph`, the intercept will always be zero.

    estimators_ : list of base learners
        The collection of fitted sub-estimators.

    train_score_ : array, shape = (n_estimators,)
        The i-th score ``train_score_[i]`` is the deviance (= loss) of the
        model at iteration ``i`` on the in-bag sample.
        If ``subsample == 1`` this is the deviance on the training data.

    oob_improvement_ : array, shape = (n_estimators,)
        The improvement in loss (= deviance) on the out-of-bag samples
        relative to the previous iteration.
        ``oob_improvement_[0]`` is the improvement in
        loss of the first stage over the ``init`` estimator.

    unique_times_ : array of shape = (n_unique_times,)
        Unique time points.

    References
    ----------
    .. [1] Hothorn, T., Bühlmann, P., Dudoit, S., Molinaro, A., van der Laan, M. J.,
           "Survival ensembles", Biostatistics, 7(3), 355-73, 2006
    """

    _tags = {
        "authors": ["sebp", "fkiraly"],  # sebp credit for interfaced estimator=
    }

    def __init__(
        self,
        loss="coxph",
        learning_rate=0.1,
        n_estimators=100,
        subsample=1.0,
        dropout_rate=0,
        verbose=0,
        random_state=None,
    ):
        self.loss = loss
        self.learning_rate = learning_rate
        self.n_estimators = n_estimators
        self.subsample = subsample
        self.dropout_rate = dropout_rate
        self.verbose = verbose
        self.random_state = random_state

        super().__init__()

    def _get_sksurv_class(self):
        """Getter of the sksurv class to be used for the adapter."""
        from sksurv.ensemble.boosting import (
            ComponentwiseGradientBoostingSurvivalAnalysis,
        )

        return ComponentwiseGradientBoostingSurvivalAnalysis

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}

        params2 = {
            "learning_rate": 0.2,
            "n_estimators": 99,
            "subsample": 0.9,
            "dropout_rate": 0.01,
        }

        return [params1, params2]



"""Cyclic boosting regressors.

This is a interface for Cyclic boosting, it contains efficient,
off-the-shelf, general-purpose supervised machine learning methods
for both regression and classification tasks.
Please read the official document for its detail
https://cyclic-boosting.readthedocs.io/en/latest/
"""

# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = [
    "setoguchi-naoki"
]  # interface only. Cyclic boosting authors in cyclic_boosting package

import warnings
from typing import Union

import numpy as np
import pandas as pd

from skpro.distributions.qpd import QPD_Johnson
from skpro.regression.base import BaseProbaRegressor


class CyclicBoosting(BaseProbaRegressor):
    """Cyclic boosting regressor from ``cyclic-boosting`` library.

    Direct interface to ``pipeline_CBAdditiveQuantileRegressor``
    and ``pipeline_CBMultiplicativeQuantileRegressor`` from ``cyclic-boosting``.

    The algorithms use boosting to create conditional distribution predictions
    that are Johnson Quantile-Parameterized Distributions (JQPD),
    with parameters estimated by quantile regression at quantile nodes.

    The quantile nodes are ``[alpha, 0.5, 1-alpha]``, where ``alpha``
    is a parameter of the model.

    The cyclic boosting model performs boosted quantile regression for the quantiles
    at the nodes, and then substitutes the quantile predictions into the paramtric
    form of the Johnson QPD.

    The model allows to select unbounded, left semi-bounded, and bounded
    predictive distribution support.

    Parameters
    ----------
    feature_groups : list, default=None
        Explanatory variables and interaction terms in the model,
        For each feature or feature tuple in the sequence, a
        one- or multidimensional factor profile will be determined,
        respectively, and used in the prediction.
        e.g. [sample1, sample2, sample3, (sample1, sample2)]
        see https://cyclic-boosting.readthedocs.io/en/latest/tutorial.html#set-features
    feature_properties : dict, default=None
        name and characteristic of train dataset by `flags` from cyclic boosting library
        it is able to set multiple characteristic by OR operator
        e.g. {sample1: flags.IS_CONTINUOUS | flags.IS_LINEAR, sample2: flags.IS_ORDERED}
        for basic options, see https://cyclic-boosting.readthedocs.io/en/latest/\
        tutorial.html#set-feature-properties
    alpha : float, default=0.2
        lower quantile QPD parameter.
        The three quantile nodes are uniquely determined by this parameter,
        as ``[alpha, 0.5, 1-alpha]``.
    mode : str, default='multiplicative'
        the type of quantile regressor. 'multiplicative' or 'additive'
    bound : str, default='U', one of ``'S'``, ``'B'``, ``'U'``
        Mode for the predictive distribution support, options are ``S``
        (semi-bounded), ``B`` (bounded), and ``U`` (unbounded).
    lower : float, default=None
        lower bound of predictive distribution support.
        If ``None`` (default), ``upper`` should also be ``None``, and the
        predictive distribution will have unbounded support, i.e., the entire reals.
        If a float, and ``upper`` is ``None``, prediction will be of
        semi-bounded support, with support between ``lower`` and infinity.
        If a float, and ``upper`` is also a float, prediction will be on a bounded
        interval, with support between ``lower`` and ``upper``.
    upper : float, default=None
        upper bound of predictive distribution support.
        If ``None`` (default), will use semi-bounded mode if ``lower`` is a float,
        and unbounded if ``lower`` is ``None``.
        If a float, assumes that ``lower`` is also a float, and prediction will
        be on a bounded interval, with support between ``lower`` and ``upper``.
    maximal_iterations : int, default=10
        maximum number of iterations for the cyclic boosting algorithm
    dist_type: str, one of ``'normal'`` (default), ``'logistic'``
        inner base distribution to use for the Johnson QPD, i.e., before
        arcosh and similar transformations.
        Available options are ``'normal'`` (default), ``'logistic'``,
        or ``'sinhlogistic'``.

    Attributes
    ----------
    estimators_ : list of skpro regressors
        clones of regressor in `estimator` fitted in the ensemble
    quantiles : list, default=[0.2, 0.5, 0.8]
        targets of quantile prediction for j-qpd's param
    quantile_values: list
        quantile prediction results
    quantile_est: list
        estimators, each estimator predicts point in the value of quantiles attribute
    qpd: skpro.distributions.J_QPD_S
        Johnson Quantile-Parameterized Distributions instance

    Examples
    --------
    >>> from skpro.regression.cyclic_boosting import CyclicBoosting
    >>> from sklearn.datasets import load_diabetes  # doctest: +SKIP
    >>> from sklearn.model_selection import train_test_split  # doctest: +SKIP
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)  # doctest: +SKIP
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)  # doctest: +SKIP

    >>> reg_proba = CyclicBoosting()  # doctest: +SKIP
    >>> reg_proba.fit(X_train, y_train)  # doctest: +SKIP
    >>> y_pred = reg_proba.predict_proba(X_test)  # doctest: +SKIP
    """

    _tags = {
        # packaging info
        # --------------
        "authors": ["setoguchi-naoki", "felix-wick"],
        "maintainers": ["setoguchi-naoki"],
        "estimator_type": "regressor_proba",
        "python_dependencies": "cyclic_boosting>=1.4.0",
        # estimator tags
        # --------------
        "capability:multioutput": False,
        "capability:missing": True,
        "X_inner_mtype": "pd_DataFrame_Table",
        "y_inner_mtype": "pd_DataFrame_Table",
    }

    def __init__(
        self,
        feature_groups=None,
        feature_properties=None,
        alpha=0.2,
        mode="multiplicative",
        lower=None,
        upper=None,
        maximal_iterations=10,
        dist_type: Union[str, None] = "normal",
        dist_shape: Union[float, None] = 0.0,
    ):
        self.feature_groups = feature_groups
        self.feature_properties = feature_properties
        self.alpha = alpha
        self.mode = mode
        self.lower = lower
        self.upper = upper
        self.maximal_iterations = maximal_iterations
        self.dist_type = dist_type
        self.dist_shape = dist_shape

        super().__init__()

        self.quantiles = [self.alpha, 0.5, 1 - self.alpha]
        self.quantile_values = list()
        self.quantile_est = list()
        self.qpd = None

        # check parameters
        if (not isinstance(feature_groups, list)) and feature_groups is not None:
            raise ValueError("feature_groups needs to be list")
        if (
            not isinstance(feature_properties, dict)
        ) and feature_properties is not None:
            raise ValueError("feature_properties needs to be dict")
        if alpha >= 0.5 or alpha <= 0.0:
            raise ValueError("alpha's range needs to be 0.0 < alpha < 0.5")

        # build estimators
        if self.mode == "multiplicative":
            from cyclic_boosting import pipeline_CBMultiplicativeQuantileRegressor

            regressor = pipeline_CBMultiplicativeQuantileRegressor
        elif self.mode == "additive":
            from cyclic_boosting import pipeline_CBAdditiveQuantileRegressor

            regressor = pipeline_CBAdditiveQuantileRegressor
        else:
            raise ValueError("mode needs to be 'multiplicative' or 'additive'")

        for quantile in self.quantiles:
            self.quantile_est.append(
                regressor(
                    quantile=quantile,
                    feature_groups=feature_groups,
                    feature_properties=feature_properties,
                    maximal_iterations=maximal_iterations,
                )
            )

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        if self.feature_groups is not None:
            feature_names = list()
            for feature in self.feature_groups:
                if isinstance(feature, tuple):
                    for f in feature:
                        feature_names.append(f)
                else:
                    feature_names.append(feature)
            if not set(feature_names).issubset(set(X.columns)):
                raise ValueError(f"{feature} is not in X")

        self._y_cols = y.columns
        y = y.to_numpy().flatten()

        # multiple quantile regression for full probability estimation
        for est in self.quantile_est:
            est.fit(X.copy(), y)

        return self

    def _predict(self, X):
        """Predict median.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        if self.feature_groups is not None:
            feature_names = list()
            for feature in self.feature_groups:
                if isinstance(feature, tuple):
                    for f in feature:
                        feature_names.append(f)
                else:
                    feature_names.append(feature)
            if not set(feature_names).issubset(set(X.columns)):
                raise ValueError(f"{feature} is not in X")

        index = X.index
        y_cols = self._y_cols

        # median prediction
        median_estimator = self.quantile_est[1]
        yhat = median_estimator.predict(X.copy())

        y_pred = pd.DataFrame(yhat, index=index, columns=y_cols)
        return y_pred

    def _predict_proba(self, X):
        """Predict QPD from three predicted quantile values.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y_pred : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        if self.feature_groups is not None:
            feature_names = list()
            for feature in self.feature_groups:
                if isinstance(feature, tuple):
                    for f in feature:
                        feature_names.append(f)
                else:
                    feature_names.append(feature)
            if not set(feature_names).issubset(set(X.columns)):
                raise ValueError(f"{feature} is not in X")

        index = X.index
        y_cols = self._y_cols

        # predict quantiles
        self.quantile_values = list()
        for est in self.quantile_est:
            yhat = est.predict(X.copy())
            self.quantile_values.append(yhat)

        # Johnson Quantile-Parameterized Distributions
        params = {
            "alpha": self.alpha,
            "qv_low": self.quantile_values[0].reshape(-1, 1),
            "qv_median": self.quantile_values[1].reshape(-1, 1),
            "qv_high": self.quantile_values[2].reshape(-1, 1),
            "lower": self.lower,
            "upper": self.upper,
            "base_dist": self.dist_type,
            "index": index,
            "columns": y_cols,
        }
        qpd = QPD_Johnson(**params)

        return qpd

    def _predict_interval(self, X, coverage):
        """Compute/return interval predictions.

        private _predict_interval containing the core logic,
            called from predict_interval and default _predict_quantiles

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        coverage : guaranteed list of float of unique values
           nominal coverage(s) of predictive interval(s)

        Returns
        -------
        pred_int : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
            second level coverage fractions for which intervals were computed,
            in the same order as in input `coverage`.
            Third level is string "lower" or "upper", for lower/upper interval end.
            Row index is equal to row index of ``X``.
            Entries are lower/upper bounds of interval predictions,
            for var in col index, at nominal coverage in second col index,
            lower/upper depending on third col index, for the row index.
            Upper/lower interval end are equivalent to
            quantile predictions at alpha = 0.5 - c/2, 0.5 + c/2 for c in coverage.
        """
        if self.feature_groups is not None:
            feature_names = list()
            for feature in self.feature_groups:
                if isinstance(feature, tuple):
                    for f in feature:
                        feature_names.append(f)
                else:
                    feature_names.append(feature)
            if not set(feature_names).issubset(set(X.columns)):
                raise ValueError(f"{feature} is not in X")

        index = X.index
        y_cols = self._y_cols
        columns = pd.MultiIndex.from_product(
            [y_cols, coverage, ["lower", "upper"]],
        )

        # predict interval
        interval = pd.DataFrame(index=index)
        for c in coverage:
            alpha = [0.5 - 0.5 * float(c), 0.5 + 0.5 * float(c)]
            interval = pd.concat(
                [interval, self.predict_quantiles(X=X.copy(), alpha=alpha)], axis=1
            )
        interval.columns = columns

        return interval

    def _predict_quantiles(self, X, alpha):
        """Compute/return quantile predictions.

        private _predict_quantiles containing the core logic,
            called from predict_quantiles and default _predict_interval

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        alpha : guaranteed list of float
            A list of probabilities at which quantile predictions are computed.

        Returns
        -------
        quantiles : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
                second level being the values of alpha passed to the function.
            Row index is equal to row index of ``X``.
            Entries are quantile predictions, for var in col index,
                at quantile probability in second col index, for the row index.
        """
        quantiles = alpha

        if self.feature_groups is not None:
            feature_names = list()
            for feature in self.feature_groups:
                if isinstance(feature, tuple):
                    for f in feature:
                        feature_names.append(f)
                else:
                    feature_names.append(feature)
            if not set(feature_names).issubset(set(X.columns)):
                raise ValueError(f"{feature} is not in X")

        is_given_proba = False
        warning = (
            "{} percentile doesn't trained, return QPD's quantile value, "
            "which is given by predict_proba(), "
            "if you need more plausible quantile value, "
            "please train regressor again for specified quantile estimation"
        )
        if isinstance(quantiles, list):
            for q in quantiles:
                if not (q in self.quantiles):
                    warnings.warn(warning.format(q), stacklevel=2)
                    is_given_proba = True
        elif isinstance(quantiles, float):
            if not (quantiles in self.quantiles):
                warnings.warn(warning.format(quantiles), stacklevel=2)
                is_given_proba = True
        else:
            raise ValueError("quantile needs to be float or list of floats")

        index = X.index
        y_cols = self._y_cols

        columns = pd.MultiIndex.from_product(
            [y_cols, quantiles],
        )

        # predict quantiles
        self.quantile_values = list()
        if is_given_proba:
            qpd = self.predict_proba(X.copy())
            pred = np.asarray([np.squeeze(qpd.ppf(q)) for q in quantiles]).T
            quantiles = pd.DataFrame(pred, index=X.index, columns=columns)

        else:
            for est in self.quantile_est:
                yhat = est.predict(X.copy())
                self.quantile_values.append(yhat)

            quantiles = pd.DataFrame(
                np.transpose(self.quantile_values), index=index, columns=columns
            )

        return quantiles

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        param1 = {
            "alpha": 0.2,
            "mode": "additive",
            "lower": 0.0,
            "maximal_iterations": 5,
        }
        param2 = {
            "alpha": 0.2,
            "mode": "additive",
            "lower": 0.0,
            "upper": 1000,
            "maximal_iterations": 5,
        }
        return [param1, param2]



"""Automated tests based on the skbase test suite template."""
import pandas as pd
import pytest
from skbase.testing import QuickTester

from skpro.datatypes import check_is_mtype, check_raise
from skpro.distributions.base import BaseDistribution
from skpro.tests.test_all_estimators import BaseFixtureGenerator, PackageConfig

TEST_ALPHAS = [0.05, [0.1], [0.25, 0.75], [0.3, 0.1, 0.9]]


class TestAllRegressors(PackageConfig, BaseFixtureGenerator, QuickTester):
    """Generic tests for all regressors in the mini package."""

    # class variables which can be overridden by descendants
    # ------------------------------------------------------

    # which object types are generated; None=all, or scitype string
    # passed to skpro.registry.all_objects as object_type
    object_type_filter = "regressor_proba"

    def test_input_output_contract(self, object_instance):
        """Tests that output of predict methods is as specified."""
        import pandas as pd
        from sklearn.datasets import load_diabetes
        from sklearn.model_selection import train_test_split

        X, y = load_diabetes(return_X_y=True, as_frame=True)
        X = X.iloc[:50]
        y = y.iloc[:50]
        y = pd.DataFrame(y)

        X_train, X_test, y_train, y_test = train_test_split(X, y)

        # fit - just once for all predict output methods
        regressor = object_instance
        regressor.fit(X_train, y_train)

        # test predict output contract
        y_pred = regressor.predict(X_test)

        assert isinstance(y_pred, pd.DataFrame)
        assert (y_pred.index == X_test.index).all()
        assert (y_pred.columns == y_train.columns).all()

        # test predict_proba output contract
        y_pred_proba = regressor.predict_proba(X_test)

        assert isinstance(y_pred_proba, BaseDistribution)
        assert (y_pred_proba.index == X_test.index).all()
        assert (y_pred_proba.columns == y_train.columns).all()
        assert y_pred_proba.shape == y_test.shape

        assert isinstance(y_pred_proba.sample(), pd.DataFrame)
        assert y_pred_proba.sample().shape == y_test.shape

        # test predict_interval output contract with default coverage
        y_pred_interval = regressor.predict_interval(X_test)

        assert isinstance(y_pred_interval, pd.DataFrame)
        assert (y_pred_interval.index == X_test.index).all()
        check_raise(
            y_pred_interval, "pred_interval", "Proba", "predict_interval return"
        )

        # test predict_quantiles output contract with default alpha
        y_pred_quantiles = regressor.predict_quantiles(X_test)

        assert isinstance(y_pred_quantiles, pd.DataFrame)
        assert (y_pred_quantiles.index == X_test.index).all()
        check_raise(
            y_pred_quantiles, "pred_quantiles", "Proba", "predict_quantiles return"
        )

        # test predict_var output contract
        y_pred_var = regressor.predict_var(X_test)

        assert isinstance(y_pred_var, pd.DataFrame)
        assert (y_pred_var.index == X_test.index).all()
        assert (y_pred_var.columns == y_train.columns).all()

    def _check_predict_quantiles(self, pred_quantiles, X_test, y_train, alpha):
        """Check expected quantile prediction output."""
        # check expected type
        valid, msg, _ = check_is_mtype(
            pred_quantiles,
            mtype="pred_quantiles",
            scitype="Proba",
            return_metadata=True,
            var_name="predict_quantiles return",
            msg_return_dict="list",
        )  # type: ignore
        assert valid, msg

        # check index
        assert (pred_quantiles.index == X_test.index).all()

        # check columns
        expected_columns = y_train.columns
        expected_quantiles = [alpha] if isinstance(alpha, float) else alpha
        expected = pd.MultiIndex.from_product([expected_columns, expected_quantiles])

        found = pred_quantiles.columns
        msg = (
            "columns of returned quantile prediction DataFrame do not"
            f"match up with expected columns. Expected: {expected},"
            f"found: {found}"
        )
        assert all(expected == found), msg

    def _check_predict_intervals(self, pred_ints, X_test, y_train, coverage):
        """Check expected interval prediction output."""
        # check expected type
        valid, msg, _ = check_is_mtype(
            pred_ints,
            mtype="pred_interval",
            scitype="Proba",
            return_metadata=True,
            var_name="predict_interval return",
            msg_return_dict="list",
        )  # type: ignore
        assert valid, msg

        # check index
        assert (pred_ints.index == X_test.index).all()

        # check columns
        expected_columns = y_train.columns
        expected_coverages = [coverage] if isinstance(coverage, float) else coverage
        expected = pd.MultiIndex.from_product(
            [expected_columns, expected_coverages, ["lower", "upper"]]
        )

        found = pred_ints.columns
        msg = (
            "columns of returned prediction interval DataFrame do not"
            f"match up with expected columns. Expected: {expected},"
            f"found: {found}"
        )
        assert all(expected == found), msg

    @pytest.mark.parametrize(
        "alpha", TEST_ALPHAS, ids=[f"alpha={a}" for a in TEST_ALPHAS]
    )
    def test_pred_quantiles_interval(self, object_instance, alpha):
        """Test predict_interval and predict_quantiles output with different alpha."""
        import pandas as pd
        from sklearn.datasets import load_diabetes
        from sklearn.model_selection import train_test_split

        X, y = load_diabetes(return_X_y=True, as_frame=True)
        X = X.iloc[:50]
        y = y.iloc[:50]
        y = pd.DataFrame(y)

        X_train, X_test, y_train, _ = train_test_split(X, y)

        regressor = object_instance
        regressor.fit(X_train, y_train)

        # check predict_interval output contract
        pred_ints = regressor.predict_interval(X_test, alpha)
        self._check_predict_intervals(pred_ints, X_test, y_train, alpha)

        # check predict_quantiles output contract
        pred_q = regressor.predict_quantiles(X_test, alpha)
        self._check_predict_quantiles(pred_q, X_test, y_train, alpha)

    def test_online_update(self, object_instance):
        """Test online update of regressor."""
        import pandas as pd
        from sklearn.datasets import load_diabetes
        from sklearn.model_selection import train_test_split

        X, y = load_diabetes(return_X_y=True, as_frame=True)
        X = X.iloc[:70]
        y = y.iloc[:70]
        y = pd.DataFrame(y)

        X_train, X_test, y_train, _ = train_test_split(X, y)
        X_fit, X_update, y_fit, y_update = train_test_split(X_train, y_train)
        X_upd1, X_upd2, y_upd1, y_upd2 = train_test_split(X_update, y_update)

        regressor = object_instance
        regressor.fit(X_fit, y_fit)

        regressor.update(X_upd1, y_upd1)
        y_pred1 = regressor.predict(X_upd1)
        y_pred2 = regressor.predict(X_upd2)

        # check predict output contract
        assert isinstance(y_pred2, pd.DataFrame)
        assert (y_pred1.index == X_upd1.index).all()
        assert (y_pred1.columns == y_fit.columns).all()
        assert (y_pred2.index == X_upd2.index).all()
        assert (y_pred2.columns == y_fit.columns).all()

        regressor.update(X_upd2, y_upd2)
        y_pred_test = regressor.predict(X_test)

        # check predict output contract
        assert isinstance(y_pred_test, pd.DataFrame)
        assert (y_pred_test.index == X_test.index).all()
        assert (y_pred_test.columns == y_fit.columns).all()



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Exponential probability distribution."""

__author__ = ["ShreeshaM07"]

import pandas as pd
from scipy.stats import expon, rv_continuous

from skpro.distributions.adapters.scipy import _ScipyAdapter


class Exponential(_ScipyAdapter):
    r"""Exponential Distribution.

    Most methods wrap ``scipy.stats.expon``.

    The Exponential distribution is parametrized by mean :math:`\mu` and
    scale :math:`b`, such that the pdf is

    .. math:: f(x) = \lambda*\exp\left(-\lambda*x\right)

    The rate :math:`\lambda` is represented by the parameter ``rate``,

    Parameter
    ---------
    rate : float or array of float (1D or 2D)
        rate of the distribution
        rate = 1/scale
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex
    """

    _tags = {
        "capabilities:approx": ["ppf", "energy", "pdfnorm"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf"],
        "distr:measuretype": "continuous",
        "broadcast_init": "on",
    }

    def __init__(self, rate, index=None, columns=None):
        self.rate = rate

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self) -> rv_continuous:
        return expon

    def _get_scipy_param(self):
        rate = self._bc_params["rate"]
        scale = 1 / rate
        return [], {"scale": scale}

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the distribution."""
        params1 = {"rate": [1, 2, 2.5, 3.5, 5]}
        params2 = {"rate": 2}
        params3 = {
            "rate": [
                [2, 2, 2],
                [4, 4, 4],
            ],
            "index": pd.Index([1, 2]),
            "columns": pd.Index(["a", "b", "c"]),
        }

        return [params1, params2, params3]



"""Base class for probabilistic regression."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

import numpy as np
import pandas as pd
from skbase.utils.dependencies import _check_estimator_deps, _check_soft_dependencies

from skpro.base import BaseEstimator
from skpro.datatypes import check_is_error_msg, check_is_mtype, convert

# allowed input mtypes
# include mtypes that are core dependencies
ALLOWED_MTYPES = [
    "pd_DataFrame_Table",
    "pd_Series_Table",
    "numpy1D",
    "numpy2D",
]
# include polars eager table if the soft dependency is installed
if _check_soft_dependencies(["polars", "pyarrow"], severity="none"):
    ALLOWED_MTYPES.append("polars_eager_table")


class BaseProbaRegressor(BaseEstimator):
    """Base class for probabilistic supervised regressors."""

    _tags = {
        "object_type": "regressor_proba",  # type of object, e.g., "distribution"
        "estimator_type": "regressor_proba",  # type of estimator, e.g., "regressor"
        "capability:survival": False,
        "capability:multioutput": False,
        "capability:missing": True,
        "capability:update": False,
        "X_inner_mtype": "pd_DataFrame_Table",
        "y_inner_mtype": "pd_DataFrame_Table",
        "C_inner_mtype": "pd_DataFrame_Table",
    }

    def __init__(self):
        super().__init__()
        _check_estimator_deps(self)

        self._X_converter_store = {}
        self._y_converter_store = {}
        self._C_converter_store = {}

    def __rmul__(self, other):
        """Magic * method, return (left) concatenated Pipeline.

        Implemented for `other` being a transformer, otherwise returns `NotImplemented`.

        Parameters
        ----------
        other: `sklearn` transformer, must follow `sklearn` API
            otherwise, `NotImplemented` is returned

        Returns
        -------
        Pipeline object,
            concatenation of `other` (first) with `self` (last).
            not nested, contains only non-Pipeline `skpro` steps
        """
        from skpro.regression.compose._pipeline import Pipeline

        # we wrap self in a pipeline, and concatenate with the other
        #   the TransformedTargetForecaster does the rest, e.g., dispatch on other
        if hasattr(other, "transform"):
            return other * Pipeline([self])
        else:
            return NotImplemented

    def fit(self, X, y, C=None):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Changes state to "fitted" = sets is_fitted flag to True

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : ignored, optional (default=None)
            censoring information for survival analysis
            All probabilistic regressors assume data to be uncensored

        Returns
        -------
        self : reference to self
        """
        capa_surv = self.get_tag("capability:survival")

        check_ret = self._check_X_y(X, y, C, return_metadata=True)

        # get inner X, y, C
        X_inner = check_ret["X_inner"]
        y_inner = check_ret["y_inner"]
        if capa_surv:
            C_inner = check_ret["C_inner"]

        # remember metadata
        self._X_metadata = check_ret["X_metadata"]
        self._y_metadata = check_ret["y_metadata"]
        if capa_surv:
            self._C_metadata = check_ret["C_metadata"]

        # set fitted flag to True
        self._is_fitted = True

        if not capa_surv:
            return self._fit(X_inner, y_inner)
        else:
            return self._fit(X_inner, y_inner, C=C_inner)

    def _fit(self, X, y, C=None):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        raise NotImplementedError

    def update(self, X, y, C=None):
        """Update regressor with a new batch of training data.

        Only estimators with the ``capability:update`` tag (value ``True``)
        provide this method, otherwise the method ignores the call and
        discards the data passed.

        State required:
            Requires state to be "fitted".

        Writes to self:
            Updates fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : ignored, optional (default=None)
            censoring information for survival analysis
            All probabilistic regressors assume data to be uncensored

        Returns
        -------
        self : reference to self
        """
        capa_online = self.get_tag("capability:update")
        capa_surv = self.get_tag("capability:survival")

        if not capa_online:
            return self

        check_ret = self._check_X_y(X, y, C, return_metadata=True)

        # get inner X, y, C
        X_inner = check_ret["X_inner"]
        y_inner = check_ret["y_inner"]
        if capa_surv:
            C_inner = check_ret["C_inner"]

        if not capa_surv:
            return self._update(X_inner, y_inner)
        else:
            return self._update(X_inner, y_inner, C=C_inner)

    def _update(self, X, y, C=None):
        """Update regressor with a new batch of training data.

        State required:
            Requires state to be "fitted".

        Writes to self:
            Updates fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        raise NotImplementedError

    def predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`
            labels predicted for `X`
        """
        X = self._check_X(X)

        y_pred = self._predict(X)

        # output conversion - back to mtype seen in fit
        y_pred = convert(
            y_pred,
            from_type=self.get_tag("y_inner_mtype"),
            to_type=self._y_metadata["mtype"],
            as_scitype="Table",
            store=self._X_converter_store,
        )

        return y_pred

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`
            labels predicted for `X`
        """
        implements_interval = self._has_implementation_of("_predict_interval")
        implements_quantiles = self._has_implementation_of("_predict_quantiles")
        implements_proba = self._has_implementation_of("_predict_proba")

        can_do_proba = implements_interval or implements_quantiles or implements_proba

        if not can_do_proba:
            raise NotImplementedError

        pred_proba = self._predict_proba(X=X)
        pred_mean = pred_proba.mean()
        return pred_mean

    def predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        X = self._check_X(X)

        y_pred = self._predict_proba(X)
        return y_pred

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        # default behaviour is implemented if one of the following three is implemented
        implements_interval = self._has_implementation_of("_predict_interval")
        implements_quantiles = self._has_implementation_of("_predict_quantiles")
        implements_var = self._has_implementation_of("_predict_var")
        can_do_proba = implements_interval or implements_quantiles or implements_var

        if not can_do_proba:
            raise NotImplementedError

        # defaulting logic is as follows:
        # var direct deputies are proba, then interval
        # proba direct deputy is var (via Normal dist)
        # quantiles direct deputies are interval, then proba
        # interval direct deputy is quantiles
        #
        # so, conditions for defaulting for proba is:
        # default to var if any of the other three are implemented

        # we use predict_var to get scale, and predict to get location
        pred_var = self.predict_var(X=X)
        pred_std = np.sqrt(pred_var)
        pred_mean = self.predict(X=X)

        from skpro.distributions.normal import Normal

        if hasattr(X, "index"):
            index = X.index
        else:
            index = pd.RangeIndex(start=0, stop=len(X), step=1)
        columns = self._get_columns(method="predict")
        pred_dist = Normal(mu=pred_mean, sigma=pred_std, index=index, columns=columns)

        return pred_dist

    def predict_interval(self, X=None, coverage=0.90):
        """Compute/return interval predictions.

        If coverage is iterable, multiple intervals will be calculated.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        coverage : float or list of float of unique values, optional (default=0.90)
           nominal coverage(s) of predictive interval(s)

        Returns
        -------
        pred_int : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
            second level coverage fractions for which intervals were computed,
            in the same order as in input `coverage`.
            Third level is string "lower" or "upper", for lower/upper interval end.
            Row index is equal to row index of ``X``.
            Entries are lower/upper bounds of interval predictions,
            for var in col index, at nominal coverage in second col index,
            lower/upper depending on third col index, for the row index.
            Upper/lower interval end are equivalent to
            quantile predictions at alpha = 0.5 - c/2, 0.5 + c/2 for c in coverage.
        """
        # check that self is fitted, if not raise exception
        self.check_is_fitted()

        # check alpha and coerce to list
        coverage = self._check_alpha(coverage, name="coverage")

        # check and convert X
        X_inner = self._check_X(X=X)

        # pass to inner _predict_interval
        pred_int = self._predict_interval(X=X_inner, coverage=coverage)
        return pred_int

    def _predict_interval(self, X, coverage):
        """Compute/return interval predictions.

        private _predict_interval containing the core logic,
            called from predict_interval and default _predict_quantiles

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        coverage : guaranteed list of float of unique values
           nominal coverage(s) of predictive interval(s)

        Returns
        -------
        pred_int : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
            second level coverage fractions for which intervals were computed,
            in the same order as in input `coverage`.
            Third level is string "lower" or "upper", for lower/upper interval end.
            Row index is equal to row index of ``X``.
            Entries are lower/upper bounds of interval predictions,
            for var in col index, at nominal coverage in second col index,
            lower/upper depending on third col index, for the row index.
            Upper/lower interval end are equivalent to
            quantile predictions at alpha = 0.5 - c/2, 0.5 + c/2 for c in coverage.
        """
        implements_quantiles = self._has_implementation_of("_predict_quantiles")
        implements_proba = self._has_implementation_of("_predict_proba")
        implements_var = self._has_implementation_of("_predict_var")
        can_do_proba = implements_quantiles or implements_proba or implements_var

        if not can_do_proba:
            raise NotImplementedError

        # defaulting logic is as follows:
        # var direct deputies are proba, then interval
        # proba direct deputy is var (via Normal dist)
        # quantiles direct deputies are interval, then proba
        # interval direct deputy is quantiles
        #
        # so, conditions for defaulting for interval are:
        # default to quantiles if any of the other three methods are implemented

        # we default to _predict_quantiles if that is implemented or _predict_proba
        # since _predict_quantiles will default to _predict_proba if it is not
        alphas = []
        for c in coverage:
            # compute quantiles corresponding to prediction interval coverage
            #  this uses symmetric predictive intervals
            alphas.extend([0.5 - 0.5 * float(c), 0.5 + 0.5 * float(c)])

        # compute quantile predictions corresponding to upper/lower
        pred_int = self._predict_quantiles(X=X, alpha=alphas)

        # change the column labels (multiindex) to the format for intervals
        # idx returned by _predict_quantiles is
        #   2-level MultiIndex with variable names, alpha
        # idx returned by _predict_interval should be
        #   3-level MultiIndex with variable names, coverage, lower/upper
        int_idx = self._get_columns(method="predict_interval", coverage=coverage)
        pred_int.columns = int_idx

        return pred_int

    def predict_quantiles(self, X=None, alpha=None):
        """Compute/return quantile predictions.

        If alpha is iterable, multiple quantiles will be calculated.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        alpha : float or list of float of unique values, optional (default=[0.05, 0.95])
            A probability or list of, at which quantile predictions are computed.

        Returns
        -------
        quantiles : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
            second level being the values of alpha passed to the function.
            Row index is equal to row index of ``X``.
            Entries are quantile predictions, for var in col index,
            at quantile probability in second col index, for the row index.
        """
        # check that self is fitted, if not raise exception
        self.check_is_fitted()

        # default alpha
        if alpha is None:
            alpha = [0.05, 0.95]
        # check alpha and coerce to list
        alpha = self._check_alpha(alpha, name="alpha")

        # input check and conversion for X
        X_inner = self._check_X(X=X)

        # pass to inner _predict_quantiles
        quantiles = self._predict_quantiles(X=X_inner, alpha=alpha)
        return quantiles

    def _predict_quantiles(self, X, alpha):
        """Compute/return quantile predictions.

        private _predict_quantiles containing the core logic,
            called from predict_quantiles and default _predict_interval

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        alpha : guaranteed list of float
            A list of probabilities at which quantile predictions are computed.

        Returns
        -------
        quantiles : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
                second level being the values of alpha passed to the function.
            Row index is equal to row index of ``X``.
            Entries are quantile predictions, for var in col index,
                at quantile probability in second col index, for the row index.
        """
        implements_interval = self._has_implementation_of("_predict_interval")
        implements_proba = self._has_implementation_of("_predict_proba")
        implements_var = self._has_implementation_of("_predict_var")
        can_do_proba = implements_interval or implements_proba or implements_var

        if not can_do_proba:
            raise NotImplementedError

        # defaulting logic is as follows:
        # var direct deputies are proba, then interval
        # proba direct deputy is var (via Normal dist)
        # quantiles direct deputies are interval, then proba
        # interval direct deputy is quantiles
        #
        # so, conditions for defaulting for quantiles are:
        # 1. default to interval if interval implemented
        # 2. default to proba if proba or var are implemented

        if implements_interval:
            pred_int = pd.DataFrame()
            for a in alpha:
                # compute quantiles corresponding to prediction interval coverage
                #  this uses symmetric predictive intervals:
                coverage = abs(1 - 2 * a)

                # compute quantile predictions corresponding to upper/lower
                pred_a = self._predict_interval(X=X, coverage=[coverage])
                pred_int = pd.concat([pred_int, pred_a], axis=1)

            # now we need to subset to lower/upper depending
            #   on whether alpha was < 0.5 or >= 0.5
            #   this formula gives the integer column indices giving lower/upper
            col_selector_int = (np.array(alpha) >= 0.5) + 2 * np.arange(len(alpha))
            col_selector_bool = np.isin(np.arange(2 * len(alpha)), col_selector_int)
            num_var = len(pred_int.columns.get_level_values(0).unique())
            col_selector_bool = np.tile(col_selector_bool, num_var)

            pred_int = pred_int.iloc[:, col_selector_bool]
            # change the column labels (multiindex) to the format for intervals
            # idx returned by _predict_interval is
            #   3-level MultiIndex with variable names, coverage, lower/upper
            # idx returned by _predict_quantiles should be
            #   is 2-level MultiIndex with variable names, alpha
            int_idx = self._get_columns(method="predict_quantiles", alpha=alpha)
            pred_int.columns = int_idx

        elif implements_proba or implements_var:
            pred_proba = self.predict_proba(X=X)
            pred_int = pred_proba.quantile(alpha=alpha)

        return pred_int

    def predict_var(self, X=None):
        """Compute/return variance predictions.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        pred_var : pd.DataFrame
            Column names are exactly those of ``y`` passed in ``fit``.
            Row index is equal to row index of ``X``.
            Entries are variance prediction, for var in col index.
            A variance prediction for given variable and fh index is a predicted
            variance for that variable and index, given observed data.
        """
        # check that self is fitted, if not raise exception
        self.check_is_fitted()

        # check and convert X
        X_inner = self._check_X(X=X)

        # pass to inner _predict_interval
        pred_var = self._predict_var(X=X_inner)
        return pred_var

    def _predict_var(self, X):
        """Compute/return variance predictions.

        private _predict_var containing the core logic, called from predict_var

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        pred_var : pd.DataFrame
            Column names are exactly those of ``y`` passed in ``fit``.
            Row index is equal to row index of ``X``.
            Entries are variance prediction, for var in col index.
            A variance prediction for given variable and fh index is a predicted
            variance for that variable and index, given observed data.
        """
        from scipy.stats import norm

        # default behaviour is implemented if one of the following three is implemented
        implements_interval = self._has_implementation_of("_predict_interval")
        implements_quantiles = self._has_implementation_of("_predict_quantiles")
        implements_proba = self._has_implementation_of("_predict_proba")
        can_do_proba = implements_interval or implements_quantiles or implements_proba

        if not can_do_proba:
            raise NotImplementedError

        # defaulting logic is as follows:
        # var direct deputies are proba, then interval
        # proba direct deputy is var (via Normal dist)
        # quantiles direct deputies are interval, then proba
        # interval direct deputy is quantiles
        #
        # so, conditions for defaulting for var are:
        # 1. default to proba if proba implemented
        # 2. default to interval if interval or quantiles are implemented

        if implements_proba:
            pred_proba = self._predict_proba(X=X)
            pred_var = pred_proba.var()
            return pred_var

        # if has one of interval/quantile predictions implemented:
        #   we get quantile prediction for first and third quartile
        #   return variance of normal distribution with that first and third quartile
        if implements_interval or implements_quantiles:
            pred_int = self._predict_interval(X=X, coverage=[0.5])
            var_names = pred_int.columns.get_level_values(0).unique()
            vars_dict = {}
            for i in var_names:
                pred_int_i = pred_int[i].copy()
                # compute inter-quartile range (IQR), as pd.Series
                iqr_i = pred_int_i.iloc[:, 1] - pred_int_i.iloc[:, 0]
                # dividing by IQR of normal gives std of normal with same IQR
                std_i = iqr_i / (2 * norm.ppf(0.75))
                # and squaring gives variance (pd.Series)
                var_i = std_i**2
                vars_dict[i] = var_i

            # put together to pd.DataFrame
            #   the indices and column names are already correct
            pred_var = pd.DataFrame(vars_dict)

        return pred_var

    def _check_X_y(self, X, y, C=None, return_metadata=False):
        X_inner, X_metadata = self._check_X(X, return_metadata=True)
        y_inner, y_metadata = self._check_y(y)

        len_X = X_metadata["n_instances"]
        len_y = y_metadata["n_instances"]

        # input check X vs y
        if len_X != "NA" and len_y != "NA" and not len_X == len_y:
            raise ValueError(
                f"X and y in fit of {self} must have same number of rows, "
                f"but X had {len_X} rows, and y had {len_y} rows"
            )

        # handle survival censoring indicator if passed and not ignored
        # only used if capability:survival tag is True
        capa_surv = self.get_tag("capability:survival")

        if capa_surv and C is not None:
            C_inner, C_metadata = self._check_C(C)
            len_C = C_metadata["n_instances"]
            if len_C != "NA" and not len_C == len_y:
                raise ValueError(
                    f"X, y, C in fit of {self} must have same number of rows, "
                    f"but C had {len_C} rows, and y had {len_y} rows"
                )
        else:
            C_inner = None
            C_metadata = None

        # in case y gets an index through conversion and X already had one
        # we need to make sure that the index of y is the same as the index of X
        # example case: X was pd.DataFrame, y was np.ndarray
        # but both get converted to X_inner, y_inner: pd.DataFrame
        # then y_inner would geta RangeIndex without this, but should have X_inner.index
        if hasattr(X_inner, "index") and not hasattr(y, "index"):
            if isinstance(y_inner, (pd.DataFrame, pd.Series)):
                y_inner.index = X_inner.index

        if hasattr(X_inner, "index") and C is not None and not hasattr(C, "index"):
            if isinstance(C_inner, (pd.DataFrame, pd.Series)):
                C_inner.index = X_inner.index

        ret_dict = {
            "X_inner": X_inner,
            "y_inner": y_inner,
        }

        if return_metadata:
            ret_dict["X_metadata"] = X_metadata
            ret_dict["y_metadata"] = y_metadata
        if capa_surv:
            ret_dict["C_inner"] = C_inner
        if return_metadata and capa_surv:
            ret_dict["C_metadata"] = C_metadata

        return ret_dict

    def _check_X(self, X, return_metadata=False):
        """Check validity of X, convert to X_inner_mtype, and return.

        Writes to self:
        feature_names_in_ = feature_names metadata of X, if not already set.
            feature_names is metadata field of X, as returned by check_is_mtype.
        n_features_in_ = number of features in X, if not already set.
            n_features_in_ = len(feature_names_in_)

        Parameters
        ----------
        X : object
            object to check and convert
        return_metadata : bool, optional, default=False
            whether to return metadata

        Returns
        -------
        X_inner : object
            X converted to X_inner_mtype = self.get_tag("X_inner_mtype")
        X_metadata : dict, only returned if return_metadata=True
            metadata of X, as returned by check_is_mtype
        """
        if return_metadata:
            req_metadata = ["n_instances", "feature_names"]
        else:
            req_metadata = ["feature_names"]
        # input validity check for X
        valid, msg, X_metadata = check_is_mtype(
            X,
            ALLOWED_MTYPES,
            "Table",
            return_metadata=req_metadata,
            var_name="X",
            msg_return_dict="list",
        )
        # shorthands for metadata used below
        X_feature_names = X_metadata["feature_names"]
        if not isinstance(X_feature_names, np.ndarray):
            X_feature_names = np.array(X_feature_names)

        # update with clearer message
        if not valid:
            check_is_error_msg(msg, var_name="X", raise_exception=True)

        # if we have seen X before, check against columns
        if hasattr(self, "feature_names_in_"):
            msg_feat = (
                f"Error in {type(self).__name__}: "
                "X in predict methods must have same columns as X in fit, "
                f"columns in fit were {self.feature_names_in_}, "
                f"but in predict found X feature names = {X_feature_names}"
            )
            if not len(X_feature_names) == len(self.feature_names_in_):
                raise ValueError(msg_feat)
            if not (X_feature_names == self.feature_names_in_).all():
                raise ValueError(msg_feat)
        # if not, remember columns
        else:
            self.feature_names_in_ = X_feature_names
            self.n_features_in_ = len(X_feature_names)

        # convert X to X_inner_mtype
        X_inner_mtype = self.get_tag("X_inner_mtype")
        X_inner = convert(
            obj=X,
            from_type=X_metadata["mtype"],
            to_type=X_inner_mtype,
            as_scitype="Table",
            store=self._X_converter_store,
        )

        if return_metadata:
            return X_inner, X_metadata
        else:
            return X_inner

    def _check_y(self, y):
        # input validity check for y
        valid, msg, metadata = check_is_mtype(
            y,
            ALLOWED_MTYPES,
            "Table",
            return_metadata=["n_instances", "feature_names"],
            var_name="y",
            msg_return_dict="list",
        )

        # update with clearer message
        if not valid:
            check_is_error_msg(msg, var_name="y", raise_exception=True)

        # convert y to y_inner_mtype
        y_inner_mtype = self.get_tag("y_inner_mtype")
        y_inner = convert(
            obj=y,
            from_type=metadata["mtype"],
            to_type=y_inner_mtype,
            as_scitype="Table",
            store=self._y_converter_store,
        )

        return y_inner, metadata

    def _check_C(self, C):
        # input validity check for C
        valid, msg, metadata = check_is_mtype(
            C,
            ALLOWED_MTYPES,
            "Table",
            return_metadata=["n_instances"],
            var_name="C",
            msg_return_dict="list",
        )

        # update with clearer message
        if not valid:
            check_is_error_msg(msg, var_name="C", raise_exception=True)

        # convert y to y_inner_mtype
        C_inner_mtype = self.get_tag("C_inner_mtype")
        C_inner = convert(
            obj=C,
            from_type=metadata["mtype"],
            to_type=C_inner_mtype,
            as_scitype="Table",
            store=self._y_converter_store,
        )

        return C_inner, metadata

    def _check_alpha(self, alpha, name="alpha"):
        """Check that quantile or confidence level value, or list of values, is valid.

        Checks:
        alpha must be a float, or list of float, all in the open interval (0, 1).
        values in alpha must be unique.

        Parameters
        ----------
        alpha : float, list of float
        name : str, optional, default="alpha"
            the name reference to alpha displayed in the error message

        Returns
        -------
        alpha coerced to a list, i.e.: [alpha], if alpha was a float; alpha otherwise

        Raises
        ------
        ValueError
            If alpha (float) or any value in alpha (list) is outside the range (0, 1).
            If values in alpha (list) are non-unique.
        """
        # check type
        if isinstance(alpha, list):
            if not all(isinstance(a, float) for a in alpha):
                raise ValueError(
                    "When `alpha` is passed as a list, it must be a list of floats"
                )

        elif isinstance(alpha, float):
            alpha = [alpha]  # make iterable

        # check range
        for a in alpha:
            if not 0 < a < 1:
                raise ValueError(
                    f"values in {name} must lie in the open interval (0, 1), "
                    f"but found value: {a}."
                )

        # check uniqueness
        if len(set(alpha)) < len(alpha):
            raise ValueError(f"values in {name} must be unique, but found duplicates")

        return alpha

    def _get_varnames(self):
        """Return variable column for DataFrame-like returns.

        Primarily used as helper for probabilistic predict-like methods.
        Assumes that _check_X_y has been called, and self._y_metadata set.

        Returns
        -------
        varnames : iterable of integer or str variable names
            can be list or pd.Index
            variable names for DataFrame-like returns
            identical to self._y_varnames if this attribute exists
        """
        featnames = self._y_metadata["feature_names"]
        return featnames

    def _get_columns(self, method="predict", **kwargs):
        """Return column names for DataFrame-like returns.

        Primarily used as helper for probabilistic predict-like methods.
        Assumes that _check_X_y has been called, and self._y_metadata set.

        Parameter
        ---------
        method : str, optional (default="predict")
            method for which to return column names
            one of "predict", "predict_interval", "predict_quantiles", "predict_var"
        kwargs : dict
            additional keyword arguments passed to private method
            important: args to private method, e.g., _predict, _predict_interval

        Returns
        -------
        columns : pd.Index
            column names
        """
        featnames = self._get_varnames()

        if method in ["predict", "predict_var"]:
            return featnames
        else:
            assert method in ["predict_interval", "predict_quantiles"]

        if method == "predict_interval":
            coverage = kwargs.get("coverage", None)
            if coverage is None:
                raise ValueError(
                    "coverage must be passed to _get_columns for predict_interval"
                )
            return pd.MultiIndex.from_product([featnames, coverage, ["lower", "upper"]])

        if method == "predict_quantiles":
            alpha = kwargs.get("alpha", None)
            if alpha is None:
                raise ValueError(
                    "alpha must be passed to _get_columns for predict_quantiles"
                )
            return pd.MultiIndex.from_product([featnames, alpha])



"""class for NGBoost probabilistic survival regression."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["ShreeshaM07"]

import numpy as np

from skpro.regression.adapters.ngboost._ngboost_proba import NGBoostAdapter
from skpro.survival.base import BaseSurvReg


class NGBoostSurvival(BaseSurvReg, NGBoostAdapter):
    """Interface of NGBSurvival of ngboost in skpro.

    NGBSurvival is a wrapper for the generic NGBoost class that
    facilitates survival analysis.
    Use this class if you want to predict an outcome that
    could take an infinite number of
    (ordered) values, but right-censoring is present in the observed data.

    Parameters
    ----------
    dist : string , default = "LogNormal"
        assumed distributional form of Y|X=x.
        A distribution from ngboost.distns, e.g. LogNormal
        Available distribution types:

        1. "LogNormal"
        2. "Exponential"

    score : string , default = "LogScore"
        rule to compare probabilistic predictions P̂ to the observed data y.
        A score from ngboost.scores, e.g. LogScore
    estimator : default learner/estimator: DecisionTreeRegressor()
        base learner to use in the boosting algorithm.
        Any instantiated sklearn regressor.
    natural_gradient : boolean , default = True
        whether natural gradient must be used or not.
    n_estimators : int , default = 500
        the number of boosting iterations to fit
    learning_rate : float , default = 0.01
        the learning rate
    minibatch_frac : float, default = 1.0
        the percent subsample of rows to
        use in each boosting iteration
    verbose : boolean, default=True
        flag indicating whether output
        should be printed during fitting
    verbose_eval : int ,default=100
        increment (in boosting iterations) at
        which output should be printed
    tol : float, default = 1e-4
        numerical tolerance to be used in optimization
    random_state : int, RandomState instance or None, optional (default=None)

    Returns
    -------
        An NGBSurvival object that can be fit.
    """

    _tags = {
        "authors": ["ShreeshaM07"],
        "maintainers": ["ShreeshaM07"],
        "python_dependencies": "ngboost",
    }

    def __init__(
        self,
        dist="LogNormal",
        score="LogScore",
        estimator=None,
        natural_gradient=True,
        n_estimators=500,
        learning_rate=0.01,
        minibatch_frac=1.0,
        col_sample=1.0,
        verbose=True,
        verbose_eval=100,
        tol=1e-4,
        random_state=None,
    ):
        self.dist = dist
        self.score = score
        self.estimator = estimator
        self.natural_gradient = natural_gradient
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.minibatch_frac = minibatch_frac
        self.col_sample = col_sample
        self.verbose = verbose
        self.verbose_eval = verbose_eval
        self.tol = tol
        self.random_state = random_state

        super().__init__()

    def _fit(self, X, y, C=None):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Changes state to "fitted" = sets is_fitted flag to True

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored

        Returns
        -------
        self : reference to self
        """
        import pandas as pd
        from ngboost import NGBSurvival
        from ngboost.scores import LogScore
        from sklearn.base import clone
        from sklearn.tree import DecisionTreeRegressor

        # skpro => 0 = uncensored, 1 = (right) censored
        # ngboost => 1 = uncensored, else (right) censored
        # If C is None then C is set as 1s (uncensored)
        # else it is converted from skpro to ngboost format
        # by doing C = 1-C
        if C is None:
            C = pd.DataFrame(np.ones(len(y)), index=y.index, columns=y.columns)
        else:
            C = 1 - C

        # coerce y to numpy array
        y = self._check_y(y=y)
        y = y[0]
        # remember y columns to predict_proba
        self._y_cols = y.columns
        y = y.values.ravel()

        if self.estimator is None:
            self.estimator_ = DecisionTreeRegressor(
                criterion="friedman_mse",
                min_samples_split=2,
                min_samples_leaf=1,
                min_weight_fraction_leaf=0.0,
                max_depth=3,
                splitter="best",
                random_state=None,
            )
        else:
            self.estimator_ = clone(self.estimator)

        dist_ngboost = self._dist_to_ngboost_instance(self.dist, survival=True)

        # Score argument for NGBSurvival
        ngboost_score = {
            "LogScore": LogScore,
        }
        score = None
        if self.score in ngboost_score:
            score = ngboost_score[self.score]

        self.ngbsurv_ = NGBSurvival(
            Dist=dist_ngboost,
            Score=score,
            Base=self.estimator_,
            natural_gradient=True,
            n_estimators=self.n_estimators,
            learning_rate=self.learning_rate,
            minibatch_frac=self.minibatch_frac,
            col_sample=self.col_sample,
            verbose=self.verbose,
            verbose_eval=self.verbose_eval,
            tol=self.tol,
            random_state=self.random_state,
        )

        # from sklearn.base import clone

        # self.ngbsurv_ = clone(self.ngbsurv)
        self.ngbsurv_.fit(X, y, C)
        return self

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        import pandas as pd

        df = pd.DataFrame(self.ngbsurv_.predict(X), index=X.index, columns=self._y_cols)
        return df

    def _pred_dist(self, X):
        return self.ngbsurv_.pred_dist(X)

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        X = self._check_X(X)

        kwargs = {}
        pred_dist = self._pred_dist(X)
        index = X.index
        columns = self._y_cols

        # Convert NGBoost Distribution return params into a dict
        kwargs = self._ngb_skpro_dist_params(pred_dist, index, columns, **kwargs)

        # Convert NGBoost Distribution to skpro BaseDistribution
        pred_dist = self._ngb_dist_to_skpro(**kwargs)

        return pred_dist

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.linear_model import LinearRegression

        params1 = {}
        params2 = {
            "dist": "LogNormal",
            "learning_rate": 0.001,
        }
        params3 = {
            "n_estimators": 800,
            "minibatch_frac": 0.8,
        }
        params4 = {
            "dist": "Exponential",
            "n_estimators": 600,
        }
        params5 = {
            "n_estimators": 200,
            "minibatch_frac": 0.8,
            "estimator": LinearRegression(),
        }

        return [params1, params2, params3, params4, params5]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Switch utility for determining whether tests for a class should be run or not.

Module does not contain tests, only test utilities.
"""

__author__ = ["fkiraly"]

from functools import lru_cache
from inspect import getmro, isclass

from skpro.tests._config import EXCLUDE_ESTIMATORS

LOCAL_PACKAGE = "skpro"


def run_test_for_class(cls, return_reason=False):
    """Check if test should run for a class or function.

    This checks the following conditions:

    1. whether all required soft dependencies are present.
       If not, does not run the test.
       If yes, behaviour depends on ONLY_CHANGED_MODULES setting:
       if off (False), always runs the test (return True);
       if on (True), runs test if and only if
       at least one of conditions 2, 3, 4 below are met.

    2. Condition 2:

      If the module containing the class/func has changed according to is_class_changed,
      or one of the modules containing any parent classes in the local package,
      then condition 2 is met.

    3. Condition 3:

      If the object is an skpro ``BaseObject``, and one of the test classes
      covering the class have changed, then condition 3 is met.

    4. Condition 4:

      If the object is an skpro ``BaseObject``, and the package requirements
      for any of its dependencies have changed in ``pyproject.toml``,
      condition 4 is met.

    cls can also be a list of classes or functions,
    in this case the test is run if and only if both of the following are True:

    * all required soft dependencies are present
    * if ``ONLY_CHANGED_MODULES`` is True, additionally,
      if any of the estimators in the list should be tested by
      at least one of criteria 2-4 above.
      If ``ONLY_CHANGED_MODULES`` is False, this condition is always True.

    Also checks whether the class or function is on the exclude override list,
    EXCLUDE_ESTIMATORS in skpro.tests._config (a list of strings, of names).
    If so, the tests are always skipped, irrespective of the other conditions.

    Parameters
    ----------
    cls : class, function or list of classes/functions
        class for which to determine whether it should be tested
    return_reason: bool, optional, default=False
        whether to return the reason for running or skipping the test

    Returns
    -------
    bool : True if class should be tested, False otherwise
        if cls was a list, is True iff True for at least one of the classes in the list
    reason: str, reason to run or skip the test, returned only if ``return_reason=True``

        * "False_exclude_list" - skip reason, class is on the exclude list
        * "False_required_deps_missing" - skip reason, required dependencies are missing
        * "False_no_change" - skip reason, no change in class or dependencies
        * "True_run_always" - run reason, run always, as ``ONLY_CHANGED_MODULES=False``
        * "True_pyproject_change" - run reason, dep(s) in ``pyproject.toml`` changed
        * "True_changed_tests" - run reason, test(s) covering class have changed
        * "True_changed_class" - run reason, module(s) containing class changed

        If multiple reasons are present, the first one in the above list is returned.

        If ``cls`` was a list, then:

        * reasons to skip - except "no change" - cause the entire list to be skipped
        * otherwise, any reasons to run cause the entire list to be run
        * otherwise, the list is not run due to "no change"
    """

    def _return(run, reason):
        if return_reason:
            return run, reason
        return run

    if isinstance(cls, (list, tuple)):
        runs = [run_test_for_class(x, return_reason=True) for x in cls]
        reasons = [x[1] for x in runs]

        # check the negative reasons that would cause the test to be skipped
        #
        # this excludes the "no change" reason, because:
        # * special negative reasons cause entire list to be skipped
        # * otherwise, positive reason causes the entire list to be run
        # * if no positive reason, then list is skipped due to lack of positive reason
        #
        # if any of the classes are on the skip list, return False
        # if any of the classes are missing dependencies, return False
        NEG_REASONS = [
            "False_exclude_list",
            "False_required_deps_missing",
        ]
        for neg_reason in NEG_REASONS:
            if any(reason == neg_reason for reason in reasons):
                return _return(False, neg_reason)

        # now check the "any of the classes should be tested" condition
        POS_REASONS = [
            "True_run_always",
            "True_pyproject_change",
            "True_changed_tests",
            "True_changed_class",
        ]
        for pos_reason in POS_REASONS:
            if any(reason == pos_reason for reason in reasons):
                return _return(True, pos_reason)

        # otherwise, we do not run, and the reason is "no change"
        return _return(False, "False_no_change")

    # if object is passed, obtain the class - objects are not hashable
    if hasattr(cls, "get_class_tag") and not isclass(cls):
        cls = cls.__class__
    # check whether estimator is on the exclude override list
    if cls.__name__ in EXCLUDE_ESTIMATORS:
        return _return(False, "False_exclude_list")

    # now we know that cls is a class or function,
    # and not on the exclude list
    run, reason = _run_test_for_class(cls)
    return _return(run, reason)


@lru_cache
def _run_test_for_class(cls):
    """Check if test should run - cached with hashable cls.

    Parameters
    ----------
    cls : class, function or list of classes/functions
        class for which to determine whether it should be tested

    Returns
    -------
    bool : True if class should be tested, False otherwise
    reason : str, reason to run or skip the test, one of:

        * "False_required_deps_missing" - skip reason, required dependencies are missing
        * "False_no_change" - skip reason, no change in class or dependencies
        * "True_run_always" - run reason, run always, as ``ONLY_CHANGED_MODULES=False``
        * "True_pyproject_change" - run reason, dep(s) in ``pyproject.toml`` changed
        * "True_changed_tests" - run reason, test(s) covering class have changed
        * "True_changed_class" - run reason, module(s) containing class changed

        If multiple reasons are present, the first one in the above list is returned.
    """
    from skbase.utils.dependencies import _check_estimator_deps

    from skpro.tests.test_all_estimators import ONLY_CHANGED_MODULES
    from skpro.utils.git_diff import get_packages_with_changed_specs, is_class_changed

    PACKAGE_REQ_CHANGED = get_packages_with_changed_specs()

    def _required_deps_present(obj):
        """Check if all required soft dependencies are present, return bool."""
        if hasattr(obj, "get_class_tag"):
            return _check_estimator_deps(obj, severity="none")
        else:
            return True

    def _is_class_changed_or_local_parents(cls):
        """Check if class or any of its local parents have changed, return bool."""
        # if cls is a function, not a class, default to is_class_changed
        if not isclass(cls):
            return is_class_changed(cls)

        # now we know cls is a class, so has an mro
        cls_and_parents = getmro(cls)
        cls_and_local_parents = [
            x for x in cls_and_parents if x.__module__.startswith(LOCAL_PACKAGE)
        ]
        return any(is_class_changed(x) for x in cls_and_local_parents)

    def _tests_covering_class_changed(cls):
        """Check if any of the tests covering cls have changed, return bool."""
        from skpro.tests.test_class_register import get_test_classes_for_obj

        test_classes = get_test_classes_for_obj(cls)
        return any(is_class_changed(x) for x in test_classes)

    def _is_impacted_by_pyproject_change(cls):
        """Check if the dep specifications of cls have changed, return bool."""
        from packaging.requirements import Requirement

        if not isclass(cls) or not hasattr(cls, "get_class_tags"):
            return False

        cls_reqs = cls.get_class_tag("python_dependencies", [])
        if cls_reqs is None:
            cls_reqs = []
        if not isinstance(cls_reqs, list):
            cls_reqs = [cls_reqs]
        package_deps = [Requirement(req).name for req in cls_reqs]

        return any(x in PACKAGE_REQ_CHANGED for x in package_deps)

    # Condition 1:
    # if any of the required soft dependencies are not present, do not run the test
    if not _required_deps_present(cls):
        return False, "False_required_deps_missing"
    # otherwise, continue

    # if ONLY_CHANGED_MODULES is off: always True
    # tests are always run if soft dependencies are present
    if not ONLY_CHANGED_MODULES:
        return True, "True_run_always"

    # run the test if and only if at least one of the conditions 2-4 are met
    # conditions are checked in order to minimize runtime due to git diff etc

    # Condition 4:
    # the package requirements for any dependency in pyproject.toml have changed
    cond4 = _is_impacted_by_pyproject_change(cls)
    if cond4:
        return True, "True_pyproject_change"

    # Condition 3:
    # if the object is an skpro BaseObject, and one of the test classes
    # covering the class have changed, then run the test
    cond3 = _tests_covering_class_changed(cls)
    if cond3:
        return True, "True_changed_tests"

    # Condition 2:
    # any of the modules containing any of the classes in the list have changed
    # or any of the modules containing any parent classes in local package have changed
    cond2 = _is_class_changed_or_local_parents(cls)
    if cond2:
        return True, "True_changed_class"

    # if none of the conditions are met, do not run the test
    # reason is that there was no change
    return False, "False_no_change"


def run_test_module_changed(module):
    """Check if test should run based on module changes

    This switch can be used to decorate tests not pertaining to a specific class.

    The function can be used to switch tests on and off
    based on whether a target module has changed.

    This checks whether the module ``module``, or any of its child modules,
    have changed.

    If ``ONLY_CHANGED_MODULES`` is False, the test is always run,
    i.e., this function always returns True.

    Parameters
    ----------
    module : string, or list of strings
        modules to check for changes, e.g., ``skpro.regression``

    Returns
    -------
    bool : switch to run or skip the test
        True iff: at least one of the modules or its submodules have changed,
        or if ``ONLY_CHANGED_MODULES`` is False
    """
    from skpro.tests.test_all_estimators import ONLY_CHANGED_MODULES
    from skpro.utils.git_diff import is_module_changed

    # if ONLY_CHANGED_MODULES is off: always True
    # tests are always run if soft dependencies are present
    if not ONLY_CHANGED_MODULES:
        return True

    if not isinstance(module, (list, tuple)):
        module = [module]

    return any(is_module_changed(mod) for mod in module)



"""Module containing accelerated failure time models."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__all__ = ["AFTFisk", "AFTLogNormal", "AFTWeibull"]

from skpro.survival.aft._aft_lifelines_fisk import AFTFisk
from skpro.survival.aft._aft_lifelines_lognormal import AFTLogNormal
from skpro.survival.aft._aft_lifelines_weibull import AFTWeibull



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Half-Normal probability distribution."""

__author__ = ["SaiRevanth25"]

import pandas as pd
from scipy.stats import halfnorm, rv_continuous

from skpro.distributions.adapters.scipy import _ScipyAdapter


class HalfNormal(_ScipyAdapter):
    r"""Half-Normal distribution.

    Most methods wrap ``scipy.stats.halfnorm``.

    This distribution is univariate, without correlation between dimensions
    for the array-valued case.

    The distribution is `cut off` at :math:`\( x = 0 \)`. There is no mass assigned to
    negative values; they are entirely excluded from the distribution.

    The half-normal distribution is parametrized by the standard deviation
    :math:`\sigma`, such that the pdf is

    .. math:: f(x) = \frac{\sqrt{2}}{\sigma \sqrt{\pi}}
                    \exp\left(-\frac{x^2}{2\sigma^2}\right), x>0 otherwise 0

    The standard deviation :math:`\sigma` is represented by the parameter ``sigma``.

    Parameters
    ----------
    sigma : float or array of float (1D or 2D), must be positive
        standard deviation of the half-normal distribution
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.halfnormal import HalfNormal

    >>> hn = HalfNormal(sigma=1)
    """

    _tags = {
        "capabilities:approx": ["pdfnorm"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, sigma, index=None, columns=None):
        self.sigma = sigma

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self) -> rv_continuous:
        return halfnorm

    def _get_scipy_param(self):
        sigma = self._bc_params["sigma"]
        return [sigma], {}

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"sigma": [[1, 2], [3, 4]]}
        params2 = {
            "sigma": 1,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"sigma": 2}
        return [params1, params2, params3]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Adapter to sklearn probabilistic regressors."""

__author__ = ["fkiraly"]

import pandas as pd

from skpro.regression.base import BaseProbaRegressor
from skpro.utils.sklearn import prep_skl_df


class SklearnProbaReg(BaseProbaRegressor):
    """Adapter to sklearn regressors with variance prediction interface.

    Wraps an sklearn regressor that can be queried for variance prediction,
    and constructs an skpro regressor from it.

    The wrapped regressor must have a ``predict`` with
    a ``return_std`` argument, and return a tuple of ``(y_pred, y_std)``,
    both ndarray of shape (n_samples,) or (n_samples, n_targets).

    Parameters
    ----------
    estimator : sklearn regressor
        Estimator to wrap, must have ``predict`` with ``return_std`` argument.
    inner_type : str, one of "pd.DataFrame", "np.ndarray", default="pd.DataFrame"
        Type of X passed to ``fit`` and ``predict`` methods of the wrapped estimator.
        Type of y passed to ``fit`` method of the wrapped estimator.
    """

    _tags = {
        "capability:multioutput": False,
        "capability:missing": True,
    }

    def __init__(self, estimator, inner_type="pd.DataFrame"):
        self.estimator = estimator
        self.inner_type = inner_type
        super().__init__()

    def _coerce_inner(self, obj):
        """Coerce obj to type of inner_type.

        Parameters
        ----------
        obj : object
            Object to coerce

        Returns
        -------
        obj : object
            Coerced object
        """
        obj = prep_skl_df(obj)
        if self.inner_type == "np.ndarray":
            obj = obj.to_numpy()
        return obj

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        from sklearn import clone

        self.estimator_ = clone(self.estimator)
        self._y_cols = y.columns

        X_inner = self._coerce_inner(X)
        y_inner = self._coerce_inner(y)

        if isinstance(y_inner, pd.DataFrame) and len(y_inner.columns) == 1:
            y_inner = y_inner.iloc[:, 0]
        elif len(y_inner.shape) > 1 and y_inner.shape[1] == 1:
            y_inner = y_inner[:, 0]

        self.estimator_.fit(X_inner, y_inner)
        return self

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        X_inner = self._coerce_inner(X)
        y_pred = self.estimator_.predict(X_inner)
        y_pred_df = pd.DataFrame(y_pred, index=X.index, columns=self._y_cols)
        return y_pred_df

    def _predict_var(self, X):
        """Compute/return variance predictions.

        private _predict_var containing the core logic, called from predict_var

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        pred_var : pd.DataFrame
            Column names are exactly those of ``y`` passed in ``fit``.
            Row index is equal to row index of ``X``.
            Entries are variance prediction, for var in col index.
            A variance prediction for given variable and fh index is a predicted
            variance for that variable and index, given observed data.
        """
        X_inner = self._coerce_inner(X)
        _, y_std = self.estimator_.predict(X_inner, return_std=True)
        y_std = pd.DataFrame(y_std, index=X.index, columns=self._y_cols)
        y_var = y_std**2
        return y_var

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.gaussian_process import GaussianProcessRegressor
        from sklearn.linear_model import BayesianRidge

        param1 = {"estimator": BayesianRidge()}
        param2 = {"estimator": GaussianProcessRegressor()}

        return [param1, param2]



"""Probabilistic regression by bootstrap."""

__author__ = ["fkiraly"]
__all__ = ["BootstrapRegressor"]

import numpy as np
import pandas as pd
from sklearn import clone
from sklearn.utils import check_random_state

from skpro.distributions.empirical import Empirical
from skpro.regression.base import BaseProbaRegressor
from skpro.utils.numpy import flatten_to_1D_if_colvector
from skpro.utils.sklearn import prep_skl_df


class BootstrapRegressor(BaseProbaRegressor):
    """Bootstrap ensemble of a tabular regressor.

    Wraps an ``sklearn`` regressor and turns it into an ``skpro`` regressor
    with access to all probabilistic prediction methods.

    Fits ``n_estimators`` clones of a tabular ``sklearn`` regressor on
    datasets which are bootstrap sub-samples, i.e.,
    independent row samples with replacement.

    On ``predict_proba``, an empirical distribution with the bootstrap
    sample is returned.

    The estimator allows to choose sample sizes for instances, variables,
    and whether sampling is with or without replacement.

    Direct generalization of ``sklearn``'s ``BaggingClassifier``
    to the probabilistic regression task.

    Parameters
    ----------
    estimator : sklearn regressor
        regressor to use in the bootstrap
    n_bootstrap_samples : int, default=100
        The number of bootstrap samples drawn
        If int, then indicates number of instances precisely
        Note: this is not the same as the size of each bootstrap sample.
        The size of the bootstrap sample is always equal to X.
    random_state : int, RandomState instance or None, optional (default=None)
        If int, ``random_state`` is the seed used by the random number generator;
        If ``RandomState`` instance, ``random_state`` is the random number generator;
        If None, the random number generator is the ``RandomState`` instance used
        by ``np.random``.

    Attributes
    ----------
    estimators_ : list of of skpro regressors
        clones of regressor in `estimator` fitted in the ensemble

    Examples
    --------
    >>> from skpro.regression.bootstrap import BootstrapRegressor
    >>> from sklearn.linear_model import LinearRegression
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.model_selection import train_test_split
    >>>
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)
    >>>
    >>> reg_tabular = LinearRegression()
    >>>
    >>> reg_proba = BootstrapRegressor(reg_tabular)
    >>> reg_proba.fit(X_train, y_train)
    BootstrapRegressor(...)
    >>> y_pred = reg_proba.predict_proba(X_test)
    """

    _tags = {"authors": "fkiraly", "capability:missing": True}

    def __init__(
        self,
        estimator,
        n_bootstrap_samples=100,
        random_state=None,
    ):
        self.estimator = estimator
        self.n_bootstrap_samples = n_bootstrap_samples
        self.random_state = random_state
        self._random_state = check_random_state(random_state)

        super().__init__()

        # todo: find the equivalent tag in sklearn for missing data handling
        # tags_to_clone = ["capability:missing"]
        # self.clone_tags(estimator, tags_to_clone)

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        estimator = self.estimator
        n_bootstrap_samples = self.n_bootstrap_samples
        np.random.seed(self.random_state)

        inst_ix = X.index
        n = len(inst_ix)

        self.estimators_ = []
        self._cols = y.columns

        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)

        for _i in range(n_bootstrap_samples):
            esti = clone(estimator)
            row_iloc = pd.RangeIndex(n)
            row_ss = _random_ss_ix(
                row_iloc, size=n, replace=True, random_state=self._random_state
            )
            inst_ix_i = inst_ix[row_ss]

            Xi = X.loc[inst_ix_i]
            Xi = Xi.reset_index(drop=True)

            yi = y.loc[inst_ix_i].values
            yi = flatten_to_1D_if_colvector(yi)

            self.estimators_ += [esti.fit(Xi, yi)]

        return self

    def _predict_proba(self, X) -> np.ndarray:
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        cols = self._cols

        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)

        y_preds = [est.predict(X) for est in self.estimators_]

        def _coerce_df(x):
            if not isinstance(x, pd.DataFrame):
                x = pd.DataFrame(x, columns=cols, index=X.index)
            return x

        y_preds = [_coerce_df(x) for x in y_preds]

        y_pred_df = pd.concat(y_preds, axis=0, keys=range(len(y_preds)))

        y_proba = Empirical(y_pred_df)
        return y_proba

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.linear_model import LinearRegression

        params1 = {"estimator": LinearRegression()}
        params2 = {
            "estimator": LinearRegression(),
            "n_bootstrap_samples": 10,
        }

        return [params1, params2]


def _random_ss_ix(ix, size, replace=True, random_state=None):
    """Randomly uniformly sample indices from a list of indices."""
    if random_state is None:
        random_state = np.random.RandomState()

    a = range(len(ix))
    ixs = ix[random_state.choice(a, size=size, replace=replace)]
    return ixs



"""Tests for scitype typing function."""

import pytest

from skpro.registry._scitype import scitype


@pytest.mark.parametrize("coerce_to_list", [True, False])
def test_scitype(coerce_to_list):
    """Test that the scitype function recovers the correct scitype(s)."""
    from skpro.distributions.laplace import Laplace
    from skpro.regression.mapie import MapieRegressor
    from skpro.regression.residual import ResidualDouble

    # test that scitype works for classes with soft dependencies
    result_mapie = scitype(MapieRegressor, coerce_to_list=coerce_to_list)
    if coerce_to_list:
        assert isinstance(result_mapie, list)
        assert "regressor_proba" == result_mapie[0]
    else:
        assert "regressor_proba" == result_mapie

    # test that scitype works for instances
    inst = ResidualDouble.create_test_instance()
    result_naive = scitype(inst, coerce_to_list=coerce_to_list)
    if coerce_to_list:
        assert isinstance(result_naive, list)
        assert "regressor_proba" == result_naive[0]
    else:
        assert "regressor_proba" == result_naive

    # test distribution object
    result_transformer = scitype(Laplace, coerce_to_list=coerce_to_list)
    if coerce_to_list:
        assert isinstance(result_transformer, list)
        assert "distribution" == result_transformer[0]
    else:
        assert "distribution" == result_transformer



"""Adapters to sklearn linear regressors with probabilistic components."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

from skpro.regression.base._delegate import _DelegatedProbaRegressor


class _DelegateWithFittedParamForwarding(_DelegatedProbaRegressor):
    """Common base class for delegates with attribute forwarding.

    Assumes that delegate has an attribute `estimator_`,
    from which fitted attributes are forwarded to self.
    """

    # attribute for _DelegatedProbaRegressor, which then delegates
    #     all non-overridden methods are same as of getattr(self, _delegate_name)
    #     see further details in _DelegatedRegressor docstring
    _delegate_name = "_estimator"
    # _estimator, not estimator_, because we do not want to expose it as
    # fitted params - fitted params are instead forwarded

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        estimator = self._get_delegate()
        estimator.fit(X=X, y=y)

        for attr in self.FITTED_PARAMS_TO_FORWARD:
            setattr(self, attr, getattr(estimator.estimator_, attr))

        return self



"""Tests for the test utilities."""

from skbase.utils.dependencies import _check_estimator_deps

from skpro.tests._config import EXCLUDE_ESTIMATORS
from skpro.tests.test_switch import run_test_for_class


def test_exclude_estimators():
    """Test that EXCLUDE_ESTIMATORS is a list of strings."""
    assert isinstance(EXCLUDE_ESTIMATORS, list)
    assert all(isinstance(estimator, str) for estimator in EXCLUDE_ESTIMATORS)


def test_run_test_for_class():
    """Test that run_test_for_class runs tests for various cases."""
    # estimator without soft deps
    from skpro.regression.bootstrap import BootstrapRegressor

    # estimator with soft deps
    from skpro.regression.mapie import MapieRegressor

    # estimator on the exception list
    from skpro.tests._config_test_dummy import DummySkipped

    # boolean flag for whether to run tests for all estimators
    from skpro.tests.test_all_estimators import ONLY_CHANGED_MODULES

    # shorthands
    f_on_excl_list = DummySkipped
    f_no_deps = BootstrapRegressor
    f_with_deps = MapieRegressor

    # test that assumptions on being on exception list are correct
    # if any of the below fail, switch the example
    assert f_on_excl_list.__name__ in EXCLUDE_ESTIMATORS
    assert f_no_deps.__name__ not in EXCLUDE_ESTIMATORS
    assert f_with_deps.__name__ not in EXCLUDE_ESTIMATORS

    # check result for skipped estimator
    run = run_test_for_class(f_on_excl_list)
    # run should be False, as the estimator is on the exception list
    assert isinstance(run, bool)
    assert not run
    # same with reason returned
    res = run_test_for_class(f_on_excl_list, return_reason=True)
    assert isinstance(res, tuple)
    assert len(res) == 2
    run, reason = res
    assert isinstance(run, bool)
    assert not run
    assert isinstance(reason, str)
    assert reason == "False_exclude_list"

    # check result for estimator without soft deps
    run = run_test_for_class(f_no_deps)
    assert isinstance(run, bool)
    if not ONLY_CHANGED_MODULES:  # if we run all tests, we should run this one
        assert run

    # result depends now on whether there is a change in the classes
    res = run_test_for_class(f_no_deps, return_reason=True)
    assert isinstance(res, tuple)
    assert len(res) == 2
    run_nodep, reason_nodep = res
    assert isinstance(run_nodep, bool)
    assert isinstance(reason_nodep, str)

    POS_REASONS = ["True_pyproject_change", "True_changed_class", "True_changed_tests"]

    if not ONLY_CHANGED_MODULES:
        assert run_nodep
        assert reason_nodep == "True_run_always"
    elif run_nodep:
        # otherwise, if we run, it must be due to changes in class or pyproject
        assert reason_nodep in POS_REASONS
    else:  # not run and only changed modules
        assert reason_nodep == "False_no_change"

    # now check estimator with soft deps
    run_wdep = run_test_for_class(f_with_deps)
    assert isinstance(run, bool)

    dep_present = _check_estimator_deps(f_with_deps, severity="none")
    if not dep_present:
        assert not run_wdep

    res = run_test_for_class(f_with_deps, return_reason=True)
    assert isinstance(res, tuple)
    assert len(res) == 2
    run_wdep, reason_wdep = res

    if not dep_present:
        assert not run_wdep
        assert reason_wdep == "False_required_deps_missing"
    elif not ONLY_CHANGED_MODULES:
        assert run_wdep
        assert reason_wdep == "True_run_always"
    elif run_wdep:
        assert reason_wdep in POS_REASONS
    else:  # not run and only changed modules
        assert reason_wdep == "False_no_change"

    # now a list of estimator with exception plus one estimator
    run = run_test_for_class([f_on_excl_list, f_no_deps])
    assert isinstance(run, bool)
    assert not run

    res = run_test_for_class([f_on_excl_list, f_no_deps], return_reason=True)
    assert isinstance(res, tuple)
    assert len(res) == 2
    run, reason = res
    assert isinstance(run, bool)
    assert not run
    assert reason == "False_exclude_list"

    # now a list of the estimator with and without soft deps
    run = run_test_for_class([f_no_deps, f_with_deps])
    assert isinstance(run, bool)

    # if deps are not present, we do not run the test
    # otherwise we run the test iff we run one of the two
    if not dep_present:
        assert not run
    else:
        assert run == run_nodep or run_wdep

    res = run_test_for_class([f_no_deps, f_with_deps], return_reason=True)
    assert isinstance(res, tuple)
    assert len(res) == 2
    run, reason = res

    if not dep_present:
        assert not run
        assert reason == "False_required_deps_missing"
    elif not ONLY_CHANGED_MODULES:
        assert run
        assert reason == "True_run_always"
    elif run:
        assert reason in POS_REASONS
        assert reason_wdep == reason or reason_nodep == reason
    else:
        assert reason == "False_no_change"
        assert reason_wdep == "False_no_change"
        assert reason_nodep == "False_no_change"
